"""
Author: Mingjian He <mh1@stanford.edu>

arn module contains autoregressive model of order n methods used in SOMATA
"""

from somata.basic_models import StateSpaceModel as Ssm
from somata.exact_inference import inverse
import numpy as np
import numbers
from scipy.linalg import block_diag


class AutoRegModel(Ssm):
    """
    AutoRegModel is a subclass of StateSpaceModel class dedicated
    to autoregressive models of order n
    """
    type = 'arn'
    default_G = None
    default_sigma2 = 3.
    order = None
    coeff = None
    sigma2 = None

    def __init__(self, coeff=None, sigma2=None,
                 components='Arn', F=None, Q=None, mu0=None, S0=None, G=None, R=None, y=None, Fs=None):
        """
        Constructor method for AutoRegModel class
        :param coeff: coefficients of AR models
        :param sigma2: noise variance of AR models
        :param components: a list of independent AR components
        :param F: transition matrix
        :param Q: state noise covariance matrix
        :param mu0: initial state mean vector
        :param S0: initial state covariance matrix
        :param G: observation matrix (row major)
        :param R: observation noise covariance matrix
        :param y: observed data (row major, can be multivariate)
        :param Fs: sampling frequency in Hz
        """
        # Autoregressive models can be constructed by directly specifying
        # the autoregressive parameters {order, coeff, sigma2}
        if coeff is not None:
            if isinstance(coeff, numbers.Number):
                coeff = (np.asanyarray([coeff], dtype=np.float64), )
            elif isinstance(coeff[0], numbers.Number):
                coeff = (np.asanyarray(coeff, dtype=np.float64),)
            else:
                coeff = tuple([np.asanyarray(x, dtype=np.float64) for x in coeff])

            order = np.asarray([len(x) for x in coeff])

            if sigma2 is not None:
                sigma2 = np.asanyarray([sigma2], dtype=np.float64) if \
                    isinstance(sigma2, numbers.Number) else np.asanyarray(sigma2, dtype=np.float64)
                assert sigma2.size == len(coeff), 'Different numbers of AR model parameters provided.'
            elif Q is not None:
                sigma2 = np.asanyarray([self._process_constructor_input(Q)[0, 0]])
            else:
                sigma2 = np.ones_like(order, dtype=np.float64) * AutoRegModel.default_sigma2

            self.order = order
            self.coeff = coeff
            self.sigma2 = sigma2

            F_tmp, Q_tmp = self._arn_to_ssm_param()
            if F is not None:
                assert (F == F_tmp).all(), 'Input state equation parameters do not agree with input F.'  # type: ignore
            if Q is not None:
                assert (Q == Q_tmp).all(), 'Input state equation parameters do not agree with input Q.'  # type: ignore
            F, Q = F_tmp, Q_tmp
        else:
            assert sigma2 is None and Q is None, 'No coefficient provided but state noise variance input is given.'

        # Provide default values for mu0 and S0
        mu0 = np.zeros((F.shape[1], 1), dtype=np.float64) if mu0 is None and F is not None else mu0
        S0 = Q[0, 0] * np.eye(Q.shape[1], dtype=np.float64) if S0 is None and Q is not None else S0

        # Fill autoregressive parameters
        self.fill_arn_param(F=F, Q=Q)

        # Set up components input to parent class constructor
        if components == 'Arn':
            if self.order is None:
                components = None
            else:
                components = []
                for ii in range(len(self.order)):
                    component = AutoRegModel()
                    component.default_G = np.hstack([np.array([[1.]], dtype=np.float64),
                                                     np.zeros((1, self.order[ii]-1), dtype=np.float64)])
                    components.append(component)
        else:
            for ii in range(len(components)):
                current_component: AutoRegModel = components[ii]  # type: ignore
                assert current_component.type == 'arn', 'Encountered non-arn type component.'
                assert current_component.default_G.shape[1] == self.order[ii], 'Components mismatch AR order.'

        # Update default_G attribute if only a single component
        if components is not None and len(components) == 1:
            self.default_G = components[0].default_G

        # Call parent class constructor
        super().__init__(components=components, F=F, Q=Q, mu0=mu0, S0=S0, G=G, R=R, y=y, Fs=Fs)

    # Dunder methods - magic methods
    def __repr__(self):
        """ Unambiguous and concise representation when calling AutoRegModel() """
        # Dynamic display depending on whether a single AR component
        if self.default_G is None:
            return super().__repr__().replace('Ssm', 'Arn')
        else:
            return 'Arn=' + str(self.default_G.shape[1]) + '<' + hex(id(self))[-4:] + '>'

    def __str__(self):
        """ Helpful information when calling print(AutoRegModel()) """
        print_str = super().__str__().replace('<Ssm object at', '<Arn object at')
        # Append additional information about autoregressive parameters
        precision_backup = np.get_printoptions()['precision']
        np.set_printoptions(precision=3)
        print_str += "{0:9} = {1}\n ".format("AR order", str(self.order))

        # create the string for displaying coeff
        if self.coeff is None:
            coeff_str = 'None'
        else:
            coeff_str = '(' + str(self.coeff[0])
            for ii in range(1, len(self.coeff)):
                coeff_str += ', ' + str(self.coeff[ii])
            coeff_str += ')'
        print_str += "{0:9} = {1}\n ".format("AR coeff", coeff_str)

        print_str += "{0:9} = {1}\n ".format("sigma2", str(self.sigma2))
        np.set_printoptions(precision=precision_backup)
        return print_str

    # Syntactic sugar methods - useful methods to make manipulations easier
    def concat_(self, other, skip_components=False):
        """
        Join two AutoRegModel objects together by concatenating the
        components.
        """
        assert self.type == 'arn', 'self in concat_() needs to be of AutoRegModel class.'

        if other.type == self.type:  # concatenation within AutoRegModel class
            # Fill autoregressive parameters in both objects first
            self.update_comp_param()
            other.update_comp_param()

            # Coefficient is a mandatory input for objects to call concat_()
            assert self.coeff is not None and other.coeff is not None, \
                'Both objects need at least coeff specified in order to concat.'
            tmp_coeff = list(self.coeff)
            for x in other.coeff:
                tmp_coeff.append(x)
            coeff = tuple(tmp_coeff)

            # Configure the rest of attributes that are immutable
            # sigma2
            if self.sigma2 is None and other.sigma2 is None:
                sigma2 = None
            elif self.sigma2 is None:
                sigma2 = np.hstack([np.ones_like(self.coeff, dtype=np.float64) * AutoRegModel.default_sigma2,
                                    other.sigma2])
            elif other.sigma2 is None:
                sigma2 = np.hstack([self.sigma2,
                                    np.ones_like(other.coeff, dtype=np.float64) * AutoRegModel.default_sigma2])
            else:
                sigma2 = np.hstack([self.sigma2, other.sigma2])

            # S0
            if self.S0 is None and other.S0 is None:
                S0 = None
            elif self.S0 is None:
                tmp_sigma2 = np.ones_like(self.order, dtype=np.float64) * AutoRegModel.default_sigma2
                _, tmp_S0 = self._arn_to_ssm_param(sigma2=tmp_sigma2)
                S0 = block_diag(tmp_S0, other.S0)
            elif other.mu0 is None:
                tmp_sigma2 = np.ones_like(other.order, dtype=np.float64) * AutoRegModel.default_sigma2
                _, tmp_S0 = AutoRegModel._arn_to_ssm_param(self=other, sigma2=tmp_sigma2)
                S0 = block_diag(self.S0, tmp_S0)
            else:
                S0 = block_diag(self.S0, other.S0)

            # Call parent class method to obtain general concatenated attributes
            temp_obj = super().concat_(other, skip_components=skip_components)
            new_obj = AutoRegModel(coeff=coeff, sigma2=sigma2, components=temp_obj.components,
                                   mu0=temp_obj.mu0, S0=S0,
                                   R=temp_obj.R, y=temp_obj.y, Fs=temp_obj.Fs)  # fill G automatically

        else:  # concatenation with other classes of SOMATA basic models
            new_obj = super().concat_(other, skip_components=skip_components)

        return new_obj

    def remove_component(self, comp_idx):
        """
        Remove a component from the AutoRegModel object,
        default to remove the left most component
        """
        if isinstance(comp_idx, list):  # recursive case
            comp_idx_list = list(np.sort(comp_idx)[::-1])
            for comp_idx in comp_idx_list:
                self.remove_component(comp_idx=comp_idx)
        else:  # base case
            super().remove_component(comp_idx=comp_idx)
            if self.order is not None:
                self.order = np.delete(self.order, comp_idx)
            if self.coeff is not None:
                self.coeff = tuple([self.coeff[x] for x in range(len(self.coeff)) if x != comp_idx])
            if self.sigma2 is not None:
                self.sigma2 = np.delete(self.sigma2, comp_idx)

            # Set attributes to default values if nothing left
            self.order = None if len(self.order) == 0 else self.order
            self.coeff = None if len(self.coeff) == 0 else self.coeff
            self.sigma2 = None if len(self.sigma2) == 0 else self.sigma2

    def fill_components(self, empty_comp=None, deep_copy=True):
        """ Fill the components attribute with AutoRegModel parameters """
        empty_comp = empty_comp or AutoRegModel()
        return super().fill_components(empty_comp=empty_comp, deep_copy=deep_copy)

    def fill_arn_param(self, F=None, Q=None):
        """ Attempt to fill autoregressive parameters """
        F = self.F if F is None else F
        Q = self.Q if Q is None else Q
        if Q is not None:
            non_zero_diagonal = np.nonzero(np.diagonal(Q))[0]
            self.order = np.asanyarray(
                np.diff(np.append(non_zero_diagonal, Q.shape[0])), dtype=np.int_)  # type: ignore
        elif F is not None:
            self.order = self._guess_ar_order(F=F)
        coeff, sigma2 = self._ssm_to_arn_param(F=F, Q=Q)
        self.coeff = coeff if coeff is not None else self.coeff
        self.sigma2 = sigma2 if sigma2 is not None else self.sigma2

    def fill_ssm_param(self, order=None, coeff=None, sigma2=None):
        """ Attempt to fill state space model parameters """
        order = self.order if order is None else order
        coeff = self.coeff if coeff is None else coeff
        sigma2 = self.sigma2 if sigma2 is None else sigma2
        F, Q = self._arn_to_ssm_param(order=order, coeff=coeff, sigma2=sigma2)
        self.F = F if F is not None else self.F
        self.Q = Q if Q is not None else self.Q

    def _guess_ar_order(self, F=None):
        """
        Try to guess the orders of AR models using F matrix,
        last resort method, should not be used normally
        """
        F = self.F if F is None else F
        assert F is not None, 'Cannot guess AR orders with None F matrix.'
        pointer = 0
        order = []
        while pointer < F.shape[0]:
            next_pointer = np.argmax(F[pointer, pointer:] == 0)
            current_order = F.shape[0] - pointer if next_pointer == 0 else next_pointer - pointer
            if current_order > 1:
                assert np.all(F[pointer+1:pointer+current_order, pointer:pointer+current_order-1] ==
                              np.eye(current_order-1, dtype=np.float64)), \
                    'Failed to guess the autoregressive model orders. Consider input order explicitly.'
            order.append(current_order)
            pointer += current_order
        return np.array(order, dtype=np.int_)

    def _ssm_to_arn_param(self, F=None, Q=None, order=None):
        """ Convert from matrices F and Q to autoregressive parameters """
        F = self.F if F is None else F
        Q = self.Q if Q is None else Q
        order = self.order if order is None else order
        if order is not None:
            coeff = []
            sigma2 = []
            for ii in range(len(order)):
                start_idx = sum(order[:ii])
                end_idx = sum(order[:ii+1])
                current_coeff = F[start_idx, start_idx:end_idx] if F is not None else None
                if isinstance(current_coeff, numbers.Number):
                    coeff.append(np.asanyarray([current_coeff], dtype=np.float64))
                else:
                    coeff.append(np.asanyarray(current_coeff, dtype=np.float64))  # type: ignore
                current_sigma2 = Q[start_idx, start_idx] if Q is not None else None  # type: ignore
                sigma2.append(current_sigma2)
            coeff = tuple(coeff) if len(order) > 1 else (coeff[0],)
            sigma2 = np.array(sigma2, dtype=np.float64)
            return coeff, sigma2
        else:
            return None, None

    def _arn_to_ssm_param(self, order=None, coeff=None, sigma2=None):
        """ Convert from autoregressive parameters to matrices F and Q """
        order = self.order if order is None else order
        coeff = self.coeff if coeff is None else coeff
        sigma2 = self.sigma2 if sigma2 is None else sigma2
        if coeff is not None:
            F_blocks = []
            Q_blocks = []
            for ii in range(len(coeff)):
                F_blocks.append(AutoRegModel.get_coeff_mat(coeff[ii]))
                Q_block = np.zeros((order[ii], order[ii]), dtype=np.float64)
                Q_block[0, 0] = sigma2[ii]
                Q_blocks.append(Q_block)
            return block_diag(*F_blocks), block_diag(*Q_blocks)
        else:
            return None, None

    def get_default_q(self, components=None, E=None):
        """
        Get the default structure of state noise covariance
        matrix Q in the Q_basis block diagonal form
        """
        components = self.components if components is None else components
        if len(components) == 1 or isinstance(components, AutoRegModel):
            order = components.default_G.shape[1] if isinstance(components, AutoRegModel) else \
                components[0].default_G.shape[1]
            E = np.eye(1, dtype=np.float64) if E is None else E
            nsource = E.shape[0]
            default_Q = np.zeros((order * nsource, order * nsource), dtype=np.float64)
            default_Q[0:nsource, 0:nsource] = E
        else:
            default_Q = block_diag(*[x.get_default_q(components=x, E=E) for x in components])
        return default_Q

    @staticmethod
    def get_coeff_mat(coeff):
        """ Create a transition matrix F from AR coefficients """
        order = len(coeff)
        return np.vstack([coeff, np.hstack([np.eye(order - 1, dtype=np.float64),
                                            np.zeros((order - 1, 1), dtype=np.float64)])])

    # Parameter estimation methods (M step)
    def m_estimate(self, **kwargs):
        """
        Maximum likelihood or Maximum a posteriori estimation to update
        parameters. Calling super().m_estimate() for all Ssm parameters
        """
        results = super().m_estimate(**kwargs)
        self.update_comp_param()
        return results

    def update_comp_param(self):
        """ Update AutoRegModel specific parameters """
        self.fill_arn_param()

    def initialize_priors(self, Q_sigma2=None, Q_hyperparameter=None,
                          R_sigma2=None, R_hyperparameter=None):
        """ Initialize priors for autoregressive models """
        # base case
        if self.ncomp <= 1:
            # [Inverse gamma prior] on state noise covariance Q non-zero diagonal entries
            if Q_sigma2 is None:
                Q_sigma2 = self.Q[0, 0]
                Q_hyperparameter = 0.1 if Q_hyperparameter is None else Q_hyperparameter

            # [Inverse gamma prior] on observation noise variance R <--- TODO: update to Wishart
            if R_sigma2 is None:
                if self.R.shape[0] > 1:
                    raise NotImplementedError('Only uni-variate observation data is supported with prior for now.')
                else:
                    R_sigma2 = self.R[0, 0]
                    R_hyperparameter = 0.1 if R_hyperparameter is None else R_hyperparameter

            return {'Q_sigma2': Q_sigma2, 'Q_hyperparameter': Q_hyperparameter,
                    'R_sigma2': R_sigma2, 'R_hyperparameter': R_hyperparameter}

        # recursive case
        else:
            assert self.components is not None, 'Cannot initialize priors outside base case when components is None.'
            components_prefill = self.fill_components()

            # expand the specified prior values to the length of components
            Q_sigma2 = self._initialize_priors_recursive_list(Q_sigma2)
            R_sigma2 = self._initialize_priors_recursive_list(R_sigma2)
            R_hyperparameter = self._initialize_priors_recursive_list(R_hyperparameter)

            # construct the final priors that is a list of dictionaries
            priors = []
            for ii in range(self.ncomp):
                current_component: AutoRegModel = self.components[ii]
                assert current_component.type == 'arn', 'Component type is not AutoRegModel class.'
                priors.append(current_component.initialize_priors(Q_sigma2=Q_sigma2[ii], R_sigma2=R_sigma2[ii],
                                                                  R_hyperparameter=R_hyperparameter[ii]))

            # unfill the components
            self.unfill_components(components_prefill)

            return priors

    @staticmethod
    def _m_update_f(A=None, B=None, priors=None):
        """ Update transition matrix -- F """
        # Update the AR coefficients -- coeff (no prior)
        approach = 'svd' if A.shape[0] >= 5 else 'gaussian'
        coeff_new = B[0, :] @ inverse(A, approach=approach)

        # Construct transition matrix -- F
        F = AutoRegModel.get_coeff_mat(coeff_new)
        return F

    @staticmethod
    def _m_update_q(A=None, B=None, C=None, T=None, F=None, priors=None):
        """ Update state noise covariance matrix -- Q """
        coeff_new = F[0, :]
        Q_ss = C[0, 0] - 2 * B[0, :] @ coeff_new + coeff_new @ A @ coeff_new

        if Ssm._m_update_if_mle('Q_sigma2', priors):
            # MLE
            sigma2_Q_new = Q_ss / T
        else:
            # MAP with inverse gamma prior
            Q_init = priors['Q_sigma2']
            Q_hp = priors['Q_hyperparameter'] if 'Q_hyperparameter' in priors else 0.1
            alpha = T * Q_hp / 2  # scales with data length T according to the hyperparameter
            beta = Q_init * (alpha + 1)  # setting the mode of inverse gamma prior to be Q_init
            sigma2_Q_new = (beta + Q_ss / 2) / (alpha + T / 2 + 1)  # mode of inverse gamma posterior

        Q = np.zeros_like(F, dtype=sigma2_Q_new.dtype)
        Q[0, 0] = sigma2_Q_new
        return Q

    @staticmethod
    def _m_update_q_src(Q_ss=None, T=None, Q_basis=None, nsource=None, nstate=None, priors=None):
        """
        Update state noise covariance matrix Q in dynamic
        source localization with a single Q_basis, in the
        Q_basis block diagonal form
        """
        import torch
        from ..source_loc import SourceLocModel as Src
        assert Q_ss.shape[0] == nsource * nstate, 'Q_ss does not match the dimension of AutoRegModel of ' \
                                                  'order ' + str(nstate) + ' across ' + str(nsource) + ' sources.'

        # Take the sums of squares from the first diagonal block
        Q_ss_block = Q_ss[0:nsource, 0:nsource]

        # Update the expansion coefficient for each source
        Q_new = torch.zeros_like(Q_ss, dtype=Q_ss.dtype).cuda()
        if isinstance(Q_basis, list):  # using non-orthonormal kernel
            Theta = Src.update_theta(Q_ss=Q_ss_block, T=T, Q_basis=Q_basis[1],
                                     nsource=nsource, npart=1, priors=priors)
            Q_new[0:nsource, 0:nsource] = Q_basis[0] @ torch.diag(Theta) @ Q_basis[0].T
        else:  # using orthonormal basis
            Theta = Src.update_theta(Q_ss=Q_ss_block, T=T, Q_basis=Q_basis,
                                     nsource=nsource, npart=1, priors=priors)
            Q_new[0:nsource, 0:nsource] = Q_basis @ torch.diag(Theta) @ Q_basis.T
        return Q_new

    @staticmethod
    def _m_update_mu0(x_0_n=None):
        """ Update initial state mean -- mu0 """
        mu0 = np.zeros((x_0_n.shape[0], 1), dtype=x_0_n.dtype)
        mu0[0, 0] = x_0_n[0]
        return mu0

    @staticmethod
    def _m_update_mu0_src(x_0_n=None, nstate=None):
        """
        Update initial state mean mu0 in dynamic
        source localization
        """
        import torch
        mu0 = torch.zeros((x_0_n.shape[0], 1), dtype=x_0_n.dtype).cuda()
        mu0[0::nstate, 0] = x_0_n[0::nstate]
        return mu0

    @staticmethod
    def _m_update_S0(x_0_n=None, P_0_n=None, mu0=None):
        """ Update initial state covariance -- S0 """
        S0 = np.zeros(P_0_n.shape, dtype=P_0_n.dtype)
        S0[0, 0] = P_0_n[0, 0] + x_0_n[0]**2 - 2 * x_0_n[0] * mu0[0, 0] + mu0[0, 0]**2
        return S0

    @staticmethod
    def _m_update_g(y=None, x_t_n=None, P_t_n=None, h_t=None, C=None, D=None):
        """ Update observation matrix -- G (AutoRegModel has fixed G) """
        return None



import numpy as np
from .multitaper_spectrogram_python import multitaper_spectrogram


def fast_psd_multitaper(x, sfreq, freq_min, freq_max, bandwidth):
    """ Fast multitaper PSD estimate with windowing """
    window_length = min(x.size / sfreq, 10)  # max to 10s windows
    time_bandwidth = np.max([1., bandwidth / 2 * window_length])
    spect, _, freq = multitaper_spectrogram(data=x, fs=sfreq, frequency_range=[freq_min, freq_max],
                                            time_bandwidth=time_bandwidth,
                                            window_params=[window_length, 1], detrend_opt='linear',
                                            multiprocess=True, plot_on=False, verbose=False)
    psd = np.mean(spect, 1) * sfreq  # scale back the normalization by sampling frequency
    return psd, freq



"""
Author: Mingjian He <mh1@stanford.edu>

basic_models module contains basic object classes defined in SOMATA
"""

from .ssm import StateSpaceModel  # Parent state-space model class  # noqa: F401
from .gen import GeneralSSModel  # General single component state-space model subclass  # noqa: F401
from .osc import OscillatorModel  # Matsuda oscillator state-space model subclass  # noqa: F401
from .arn import AutoRegModel  # Autoregressive model of order n subclass  # noqa: F401



"""
Author: Mingjian He <mh1@stanford.edu>

em module contains a general run_em() function that works across SOMATA
"""

import numpy as np
from tqdm import tqdm


def run_em(obj: object, y=None, init_from_data=False, e_kwargs=None, m_kwargs=None,
           max_iter=10, stop_thresh=np.finfo(float).eps, ignore_numerr=False, return_dict=False, show_pbar=False):
    """
    run_em() is a general purpose function to organize EM algorithms used
    throughout SOMATA. The class object that is input as the first argument
    is expected to have e_step() and m_step() properly defined. Additional
    inputs to the E step and M step can be provided as dictionaries for
    e_kwargs and m_kwargs to make further specifications such as priors.
    This function is implemented here as a high-level organizer of EM
    iterations to bring together different signal processing algorithms
    that all depend on the generalized EM iteration structure, despite
    their specific E, M steps can be vastly different.

    Inputs:
    :param obj: a SOMATA-compatible class object, could be one of basic_models
                objects, or a module specific class object
    :param y: observed data
    :param init_from_data: boolean flag to call initialize_from_data() method on obj
    :param e_kwargs: arguments for exposed e_step() method
    :param m_kwargs: arguments for exposed m_step() method
    :param max_iter: maximal number of EM iterations to run
    :param stop_thresh: a stopping criterion threshold, flexible depending on the
                        stopping variable used by the e/m_step() methods of obj
    :param ignore_numerr: whether to ignore warning messages of numerical errors
    :param return_dict: None -> no return, True -> return dict, False -> return tuple of variables
    :param show_pbar: show progress bar during EM iterations
    """
    e_kwargs = e_kwargs or {}
    m_kwargs = m_kwargs or {}

    # Initialize from data
    if init_from_data:
        obj.initialize_from_data(y=y)

    # Initialize EM iterations
    em_iter = 0
    stop_var = float('inf')
    stop_var_tally = []

    if ignore_numerr:
        old_settings = np.seterr(divide='ignore', over='ignore', invalid='ignore')

    # Run EM to learn model parameters
    with tqdm(total=max_iter, desc='run_em()', colour='green', disable=not show_pbar) as pbar:
        while stop_var > stop_thresh and em_iter < max_iter:
            # E step
            e_results, stop_var = obj.e_step(y=y, **e_kwargs)
            stop_var_tally.append(stop_var)
            if isinstance(stop_var, bool) and stop_var:
                break  # break EM while loop
            else:
                assert stop_var >= 0, 'Stopping variable should be non-negative. Current value = ' + str(stop_var) + '!'

            # M step
            obj.m_step(y=y, **e_results, **m_kwargs)

            em_iter += 1
            pbar.update(1)

    if ignore_numerr:
        # noinspection PyUnboundLocalVariable
        np.seterr(**old_settings)

    if return_dict is None:
        pass
    elif return_dict:
        return {'em_iter': em_iter, 'stop_var_tally': stop_var_tally}
    else:
        return em_iter, stop_var_tally



"""
Author: Mingjian He <mh1@stanford.edu>

switching module contains switching state-space models in SOMATA
"""

from .traditional import switching  # noqa: F401
from .vb import VBSwitchModel  # noqa: F401



# Analysis Imports
import math
import numpy as np
from scipy.signal.windows import dpss
from scipy.signal import detrend
# Logistical Imports
import warnings
import timeit
from joblib import Parallel, delayed, cpu_count
# Visualization imports
import colorcet  # this import is necessary to add rainbow colormap to matplotlib  # noqa: F401
import matplotlib.pyplot as plt


# MULTITAPER SPECTROGRAM #
def multitaper_spectrogram(data, fs, frequency_range=None, time_bandwidth=5, num_tapers=None, window_params=None,
                           min_nfft=0, detrend_opt='linear', multiprocess=False, n_jobs=None, weighting='unity',
                           plot_on=True, return_fig=False, clim_scale=True, verbose=True, xyflip=False, ax=None):
    """ Compute multitaper spectrogram of timeseries data
    Usage:
    mt_spectrogram, stimes, sfreqs = multitaper_spectrogram(data, fs, frequency_range=None, time_bandwidth=5,
                                                            num_tapers=None, window_params=None, min_nfft=0,
                                                            detrend_opt='linear', multiprocess=False, cpus=False,
                                                            weighting='unity', plot_on=True, return_fig=False,
                                                            clim_scale=True, verbose=True, xyflip=False):
        Arguments:
                data (1d np.array): time series data -- required
                fs (float): sampling frequency in Hz  -- required
                frequency_range (list): 1x2 list - [<min frequency>, <max frequency>] (default: [0 nyquist])
                time_bandwidth (float): time-half bandwidth product (window duration*half bandwidth of main lobe)
                                        (default: 5 Hz*s)
                num_tapers (int): number of DPSS tapers to use (default: [will be computed
                                  as floor(2*time_bandwidth - 1)])
                window_params (list): 1x2 list - [window size (seconds), step size (seconds)] (default: [5 1])
                detrend_opt (string): detrend data window ('linear' (default), 'constant', 'off')
                                      (Default: 'linear')
                min_nfft (int): minimum allowable NFFT size, adds zero padding for interpolation (closest 2^x)
                                (default: 0)
                multiprocess (bool): Use multiprocessing to compute multitaper spectrogram (default: False)
                n_jobs (int): Number of cpus to use if multiprocess = True (default: False). Note: if default is left
                            as None and multiprocess = True, the number of cpus used for multiprocessing will be
                            all available - 1.
                weighting (str): weighting of tapers ('unity' (default), 'eigen', 'adapt');
                plot_on (bool): plot results (default: True)
                return_fig (bool): return plotted spectrogram (default: False)
                clim_scale (bool): automatically scale the colormap on the plotted spectrogram (default: True)
                verbose (bool): display spectrogram properties (default: True)
                xyflip (bool): transpose the mt_spectrogram output (default: False)
                ax (axes): a matplotlib axes to plot the spectrogram on (default: None)
        Returns:
                mt_spectrogram (TxF np array): spectral power matrix
                stimes (1xT np array): timepoints (s) in mt_spectrogram
                sfreqs (1xF np array)L frequency values (Hz) in mt_spectrogram

        Example:
        In this example we create some chirp data and run the multitaper spectrogram on it.
            import numpy as np  # import numpy
            from scipy.signal import chirp  # import chirp generation function
            # Set spectrogram params
            fs = 200  # Sampling Frequency
            frequency_range = [0, 25]  # Limit frequencies from 0 to 25 Hz
            time_bandwidth = 3  # Set time-half bandwidth
            num_tapers = 5  # Set number of tapers (optimal is time_bandwidth*2 - 1)
            window_params = [4, 1]  # Window size is 4s with step size of 1s
            min_nfft = 0  # No minimum nfft
            detrend_opt = 'constant'  # detrend each window by subtracting the average
            multiprocess = True  # use multiprocessing
            cpus = 3  # use 3 cores in multiprocessing
            weighting = 'unity'  # weight each taper at 1
            plot_on = True  # plot spectrogram
            return_fig = False  # do not return plotted spectrogram
            clim_scale = False # don't auto-scale the colormap
            verbose = True  # print extra info
            xyflip = False  # do not transpose spect output matrix

            # Generate sample chirp data
            t = np.arange(1/fs, 600, 1/fs)  # Create 10 min time array from 1/fs to 600 stepping by 1/fs
            f_start = 1  # Set chirp freq range min (Hz)
            f_end = 20  # Set chirp freq range max (Hz)
            data = chirp(t, f_start, t[-1], f_end, 'logarithmic')
            # Compute the multitaper spectrogram
            spect, stimes, sfreqs = multitaper_spectrogram(data, fs, frequency_range, time_bandwidth, num_tapers,
                                                           window_params, min_nfft, detrend_opt, multiprocess,
                                                           cpus, weighting, plot_on, return_fig, clim_scale,
                                                           verbose, xyflip):

        This code is companion to the paper:
        "Sleep Neurophysiological Dynamics Through the Lens of Multitaper Spectral Analysis"
           Michael J. Prerau, Ritchie E. Brown, Matt T. Bianchi, Jeffrey M. Ellenbogen, Patrick L. Purdon
           December 7, 2016 : 60-92
           DOI: 10.1152/physiol.00062.2015
         which should be cited for academic use of this code.

         A full tutorial on the multitaper spectrogram can be found at: # https://www.sleepEEG.org/multitaper

        Copyright 2021 Michael J. Prerau Laboratory. - https://www.sleepEEG.org
        Authors: Michael J. Prerau, Ph.D., Thomas Possidente, Mingjian He

  __________________________________________________________________________________________________________________
    """

    #  Process user input
    [data, fs, frequency_range, time_bandwidth, num_tapers,
     winsize_samples, winstep_samples, window_start,
     num_windows, nfft, detrend_opt, plot_on, verbose] = process_input(data, fs, frequency_range, time_bandwidth,
                                                                       num_tapers, window_params, min_nfft,
                                                                       detrend_opt, plot_on, verbose)

    # Set up spectrogram parameters
    [window_idxs, stimes, sfreqs, freq_inds] = process_spectrogram_params(fs, nfft, frequency_range, window_start,
                                                                          winsize_samples)
    # Display spectrogram parameters
    if verbose:
        display_spectrogram_props(fs, time_bandwidth, num_tapers, [winsize_samples, winstep_samples], frequency_range,
                                  nfft, detrend_opt)

    # Split data into segments and preallocate
    data_segments = data[window_idxs]

    # COMPUTE THE MULTITAPER SPECTROGRAM
    #     STEP 1: Compute DPSS tapers based on desired spectral properties
    #     STEP 2: Multiply the data segment by the DPSS Tapers
    #     STEP 3: Compute the spectrum for each tapered segment
    #     STEP 4: Take the mean of the tapered spectra

    # Compute DPSS tapers (STEP 1)
    dpss_tapers, dpss_eigen = dpss(winsize_samples, time_bandwidth, num_tapers, return_ratios=True)
    dpss_eigen = np.reshape(dpss_eigen, (num_tapers, 1))

    # pre-compute weights
    if weighting == 'eigen':
        wt = dpss_eigen / num_tapers
    elif weighting == 'unity':
        wt = np.ones(num_tapers) / num_tapers
        wt = np.reshape(wt, (num_tapers, 1))  # reshape as column vector
    else:
        wt = 0

    tic = timeit.default_timer()  # start timer

    # Set up calc_mts_segment() input arguments
    mts_params = (dpss_tapers, nfft, freq_inds, detrend_opt, num_tapers, dpss_eigen, weighting, wt)

    if multiprocess:  # use multiprocessing
        n_jobs = max(cpu_count() - 1, 1) if n_jobs is None else n_jobs
        mt_spectrogram = np.vstack(Parallel(n_jobs=n_jobs)(delayed(calc_mts_segment)(
            data_segments[num_window, :], *mts_params) for num_window in range(num_windows)))

    else:  # if no multiprocessing, compute normally
        mt_spectrogram = np.apply_along_axis(calc_mts_segment, 1, data_segments, *mts_params)

    # Compute one-sided PSD spectrum
    mt_spectrogram = mt_spectrogram.T
    dc_select = np.where(sfreqs == 0)[0]
    nyquist_select = np.where(sfreqs == fs/2)[0]
    select = np.setdiff1d(np.arange(0, len(sfreqs)), np.concatenate((dc_select, nyquist_select)))

    mt_spectrogram = np.vstack([mt_spectrogram[dc_select, :], 2*mt_spectrogram[select, :],
                               mt_spectrogram[nyquist_select, :]]) / fs

    # Flip if requested
    if xyflip:
        mt_spectrogram = mt_spectrogram.T

    # End timer and get elapsed compute time
    toc = timeit.default_timer()
    if verbose:
        print("\n Multitaper compute time: " + "%.2f" % (toc - tic) + " seconds")

    if np.all(mt_spectrogram.flatten() == 0):
        print("\n Data was all zeros, no output")

    # Plot multitaper spectrogram
    if plot_on:
        # convert from power to dB
        spect_data = nanpow2db(mt_spectrogram)

        # Set x and y axes
        dx = stimes[1] - stimes[0]
        dy = sfreqs[1] - sfreqs[0]
        extent = [stimes[0]-dx, stimes[-1]+dx, sfreqs[-1]+dy, sfreqs[0]-dy]

        # Plot spectrogram
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()
        im = ax.imshow(spect_data, extent=extent, aspect='auto')
        fig.colorbar(im, ax=ax, label='PSD (dB)', shrink=0.8)
        ax.set_xlabel("Time (HH:MM:SS)")
        ax.set_ylabel("Frequency (Hz)")
        im.set_cmap(plt.cm.get_cmap('cet_rainbow4'))
        ax.invert_yaxis()

        # Scale colormap
        if clim_scale:
            clim = np.percentile(spect_data, [5, 98])  # from 5th percentile to 98th
            im.set_clim(clim)  # actually change colorbar scale

        if return_fig:
            return mt_spectrogram, stimes, sfreqs, (fig, ax)

    return mt_spectrogram, stimes, sfreqs


# Helper Functions #

# Process User Inputs #
def process_input(data, fs, frequency_range=None, time_bandwidth=5, num_tapers=None, window_params=None, min_nfft=0,
                  detrend_opt='linear', plot_on=True, verbose=True):
    """ Helper function to process multitaper_spectrogram() arguments
            Arguments:
                    data (1d np.array): time series data-- required
                    fs (float): sampling frequency in Hz  -- required
                    frequency_range (list): 1x2 list - [<min frequency>, <max frequency>] (default: [0 nyquist])
                    time_bandwidth (float): time-half bandwidth product (window duration*half bandwidth of main lobe)
                                            (default: 5 Hz*s)
                    num_tapers (int): number of DPSS tapers to use (default: None [will be computed
                                      as floor(2*time_bandwidth - 1)])
                    window_params (list): 1x2 list - [window size (seconds), step size (seconds)] (default: [5 1])
                    min_nfft (int): minimum allowable NFFT size, adds zero padding for interpolation (closest 2^x)
                                    (default: 0)
                    detrend_opt (string): detrend data window ('linear' (default), 'constant', 'off')
                                          (Default: 'linear')
                    plot_on (True): plot results (default: True)
                    verbose (True): display spectrogram properties (default: true)
            Returns:
                    data (1d np.array): same as input
                    fs (float): same as input
                    frequency_range (list): same as input or calculated from fs if not given
                    time_bandwidth (float): same as input or default if not given
                    num_tapers (int): same as input or calculated from time_bandwidth if not given
                    winsize_samples (int): number of samples in single time window
                    winstep_samples (int): number of samples in a single window step
                    window_start (1xm np.array): array of timestamps representing the beginning time for each window
                    num_windows (int): number of windows in the data
                    nfft (int): length of signal to calculate fft on
                    detrend_opt ('string'): same as input or default if not given
                    plot_on (bool): same as input
                    verbose (bool): same as input
    """

    # Make sure data is 1 dimensional np array
    if len(data.shape) != 1:
        if (len(data.shape) == 2) & (data.shape[1] == 1):  # if it's 2d, but can be transferred to 1d, do so
            data = np.ravel(data[:, 0])
        elif (len(data.shape) == 2) & (data.shape[0] == 1):  # if it's 2d, but can be transferred to 1d, do so
            data = np.ravel(data.T[:, 0])
        else:
            raise TypeError("Input data is the incorrect dimensions. Should be a 1d array with shape (n,) where n is \
                            the number of data points. Instead data shape was " + str(data.shape))

    # Set frequency range if not provided
    if frequency_range is None:
        frequency_range = [0, fs / 2]

    # Set detrending method
    detrend_opt = detrend_opt.lower()
    if detrend_opt != 'linear':
        if detrend_opt in ['const', 'constant']:
            detrend_opt = 'constant'
        elif detrend_opt in ['none', 'false', 'off']:
            detrend_opt = 'off'
        else:
            raise ValueError("'" + str(detrend_opt) + "' is not a valid argument for detrend_opt. The choices " +
                             "are: 'constant', 'linear', or 'off'.")
    # Check if frequency range is valid
    if frequency_range[1] > fs / 2:
        frequency_range[1] = fs / 2
        warnings.warn('Upper frequency range greater than Nyquist, setting range to [' +
                      str(frequency_range[0]) + ', ' + str(frequency_range[1]) + ']')

    # Set number of tapers if none provided
    if num_tapers is None:
        num_tapers = math.floor(2 * time_bandwidth) - 1

    # Warn if number of tapers is suboptimal
    if num_tapers != math.floor(2 * time_bandwidth) - 1:
        warnings.warn('Number of tapers is optimal at floor(2*TW) - 1. consider using ' +
                      str(math.floor(2 * time_bandwidth) - 1))

    # If no window params provided, set to defaults
    if window_params is None:
        window_params = [5, 1]

    # Check if window size is valid, fix if not
    if window_params[0] * fs % 1 != 0:
        winsize_samples = round(window_params[0] * fs)
        warnings.warn('Window size is not divisible by sampling frequency. Adjusting window size to ' +
                      str(winsize_samples / fs) + ' seconds')
    else:
        winsize_samples = window_params[0] * fs

    # Check if window step is valid, fix if not
    if window_params[1] * fs % 1 != 0:
        winstep_samples = round(window_params[1] * fs)
        warnings.warn('Window step size is not divisible by sampling frequency. Adjusting window step size to ' +
                      str(winstep_samples / fs) + ' seconds')
    else:
        winstep_samples = window_params[1] * fs

    # Get total data length
    len_data = len(data)

    # Check if length of data is smaller than window (bad)
    if len_data < winsize_samples:
        raise ValueError("\nData length (" + str(len_data) + ") is shorter than window size (" +
                         str(winsize_samples) + "). Either increase data length or decrease window size.")

    # Find window start indices and num of windows
    window_start = np.arange(0, len_data - winsize_samples + 1, winstep_samples)
    num_windows = len(window_start)

    # Get num points in FFT
    if min_nfft == 0:  # avoid divide by zero error in np.log2(0)
        nfft = max(2 ** math.ceil(np.log2(abs(winsize_samples))), winsize_samples)
    else:
        nfft = max(max(2 ** math.ceil(np.log2(abs(winsize_samples))), winsize_samples),
                   2 ** math.ceil(np.log2(abs(min_nfft))))

    return ([data, fs, frequency_range, time_bandwidth, num_tapers,
             int(winsize_samples), int(winstep_samples), window_start, num_windows, nfft,
             detrend_opt, plot_on, verbose])


# PROCESS THE SPECTROGRAM PARAMETERS #
def process_spectrogram_params(fs, nfft, frequency_range, window_start, datawin_size):
    """ Helper function to create frequency vector and window indices
        Arguments:
             fs (float): sampling frequency in Hz  -- required
             nfft (int): length of signal to calculate fft on -- required
             frequency_range (list): 1x2 list - [<min frequency>, <max frequency>] -- required
             window_start (1xm np array): array of timestamps representing the beginning time for each
                                          window -- required
             datawin_size (float): seconds in one window -- required
        Returns:
            window_idxs (nxm np array): indices of timestamps for each window
                                        (nxm where n=number of windows and m=datawin_size)
            stimes (1xt np array): array of times for the center of the spectral bins
            sfreqs (1xf np array): array of frequency bins for the spectrogram
            freq_inds (1d np array): boolean array of which frequencies are being analyzed in
                                      an array of frequencies from 0 to fs with steps of fs/nfft
    """

    # create frequency vector
    df = fs / nfft
    sfreqs = np.arange(0, fs, df)

    # Get frequencies for given frequency range
    freq_inds = (sfreqs >= frequency_range[0]) & (sfreqs <= frequency_range[1])
    sfreqs = sfreqs[freq_inds]

    # Compute times in the middle of each spectrum
    window_middle_samples = window_start + round(datawin_size / 2)
    stimes = window_middle_samples / fs

    # Get indexes for each window
    window_idxs = np.atleast_2d(window_start).T + np.arange(0, datawin_size, 1)
    window_idxs = window_idxs.astype(int)

    return [window_idxs, stimes, sfreqs, freq_inds]


# DISPLAY SPECTROGRAM PROPERTIES
def display_spectrogram_props(fs, time_bandwidth, num_tapers, data_window_params, frequency_range, nfft, detrend_opt):
    """ Prints spectrogram properties
        Arguments:
            fs (float): sampling frequency in Hz  -- required
            time_bandwidth (float): time-half bandwidth product (window duration*1/2*frequency_resolution) -- required
            num_tapers (int): number of DPSS tapers to use -- required
            data_window_params (list): 1x2 list - [window length(s), window step size(s)] -- required
            frequency_range (list): 1x2 list - [<min frequency>, <max frequency>] -- required
            nfft(float): number of fast fourier transform samples -- required
            detrend_opt (str): detrend data window ('linear' (default), 'constant', 'off') -- required
        Returns:
            This function does not return anything
    """

    data_window_params = np.asarray(data_window_params) / fs

    # Print spectrogram properties
    print("Multitaper Spectrogram Properties: ")
    print('     Spectral Resolution: ' + str(2 * time_bandwidth / data_window_params[0]) + 'Hz')
    print('     Window Length: ' + str(data_window_params[0]) + 's')
    print('     Window Step: ' + str(data_window_params[1]) + 's')
    print('     Time Half-Bandwidth Product: ' + str(time_bandwidth))
    print('     Number of Tapers: ' + str(num_tapers))
    print('     Frequency Range: ' + str(frequency_range[0]) + "-" + str(frequency_range[1]) + 'Hz')
    print('     NFFT: ' + str(nfft))
    print('     Detrend: ' + detrend_opt + '\n')


# NANPOW2DB
def nanpow2db(y):
    """ Power to dB conversion, setting bad values to nans
        Arguments:
            y (float or array-like): power
        Returns:
            ydB (float or np array): inputs converted to dB with 0s and negatives resulting in nans
    """

    if isinstance(y, int) or isinstance(y, float):
        if y == 0:
            return np.nan
        else:
            ydB = 10 * np.log10(y)
    else:
        if isinstance(y, list):  # if list, turn into array
            y = np.asarray(y)
        y = y.astype(float)  # make sure it's a float array so we can put nans in it
        y[y == 0] = np.nan
        ydB = 10 * np.log10(y)

    return ydB


# Helper #
def is_outlier(data):
    smad = 1.4826 * np.median(abs(data - np.median(data)))  # scaled median absolute deviation
    outlier_mask = abs(data-np.median(data)) > 3*smad  # outliers are more than 3 smads away from median
    outlier_mask = (outlier_mask | np.isnan(data) | np.isinf(data))
    return outlier_mask


# CALCULATE MULTITAPER SPECTRUM ON SINGLE SEGMENT
def calc_mts_segment(data_segment, dpss_tapers, nfft, freq_inds, detrend_opt, num_tapers, dpss_eigen, weighting, wt):
    """ Helper function to calculate the multitaper spectrum of a single segment of data
        Arguments:
            data_segment (1d np.array): One window worth of time-series data -- required
            dpss_tapers (2d np.array): Parameters for the DPSS tapers to be used.
                                       Dimensions are (num_tapers, winsize_samples) -- required
            nfft (int): length of signal to calculate fft on -- required
            freq_inds (1d np array): boolean array of which frequencies are being analyzed in
                                      an array of frequencies from 0 to fs with steps of fs/nfft
            detrend_opt (str): detrend data window ('linear' (default), 'constant', 'off')
            num_tapers (int): number of tapers being used
            dpss_eigen (np array):
            weighting (str):
            wt (int or np array):
        Returns:
            mt_spectrum (1d np.array): spectral power for single window
    """

    # If segment has all zeros, return vector of zeros
    if all(data_segment == 0):
        ret = np.empty(sum(freq_inds))
        ret.fill(0)
        return ret

    if any(np.isnan(data_segment)):
        ret = np.empty(sum(freq_inds))
        ret.fill(np.nan)
        return ret

    # Option to detrend data to remove low frequency DC component
    if detrend_opt != 'off':
        data_segment = detrend(data_segment, type=detrend_opt)

    # Multiply data by dpss tapers (STEP 2)
    tapered_data = np.multiply(np.mat(data_segment).T, np.mat(dpss_tapers.T))

    # Compute the FFT (STEP 3)
    fft_data = np.fft.fft(tapered_data, nfft, axis=0)

    # Compute the weighted mean spectral power across tapers (STEP 4)
    spower = np.power(np.imag(fft_data), 2) + np.power(np.real(fft_data), 2)
    if weighting == 'adapt':
        # adaptive weights - for colored noise spectrum (Percival & Walden p368-370)
        tpower = np.dot(np.transpose(data_segment), (data_segment/len(data_segment)))
        spower_iter = np.mean(spower[:, 0:2], 1)
        spower_iter = spower_iter[:, np.newaxis]
        a = (1 - dpss_eigen) * tpower
        for i in range(3):  # 3 iterations only
            # Calc the MSE weights
            b = np.dot(spower_iter, np.ones((1, num_tapers))) / ((np.dot(spower_iter, np.transpose(dpss_eigen))) +
                                                                 (np.ones((nfft, 1)) * np.transpose(a)))
            # Calc new spectral estimate
            wk = (b**2) * np.dot(np.ones((nfft, 1)), np.transpose(dpss_eigen))
            spower_iter = np.sum((np.transpose(wk) * np.transpose(spower)), 0) / np.sum(wk, 1)
            spower_iter = spower_iter[:, np.newaxis]

        mt_spectrum = np.squeeze(spower_iter)

    else:
        # eigenvalue or uniform weights
        mt_spectrum = np.dot(spower, wt)
        mt_spectrum = np.reshape(mt_spectrum, nfft)  # reshape to 1D

    return mt_spectrum[freq_inds]



"""
Author: Mingjian He <mh1@stanford.edu>

Dynamic programming (also known as belief propagation) algorithms for exact inference
"""

import torch
import numpy as np
from scipy.linalg import solve_triangular


def forward_backward(A, py_x, p1=None, compute_edge=False):
    """
    Classical forward-backward algorithm for discrete state HMM with
    step-wise normalization. Note that this ensures the final conditional
    densities are correct, but the alpha and beta vectors are off by a
    normalization constant

    Reference:
        Rabiner, L., & Juang, B. (1986). An introduction to hidden
        Markov models. ieee assp magazine, 3(1), 4-16.

    Inputs:
    :param A: transition probability matrix
    :param py_x: observation probability
    :param p1: initial prior of hidden state at t=1
    :param compute_edge: whether to compute edge marginals
    """
    # Model dimensions
    K = A.shape[0]
    T = py_x.shape[1]
    assert py_x.shape[0] == K, 'Dimension mismatch between A and py_x.'
    logL = np.zeros(T)

    # Initialize prior at t=1
    p1 = np.ones(K) / K if p1 is None else p1

    # Compute the alpha (forward pass)
    norm_a = np.zeros((K, T))
    norm_a[:, 0] = py_x[:, 0] * p1 / (py_x[:, 0] * p1).sum()  # t=1
    logL[0] = np.log((py_x[:, 0] * p1).sum())
    for ii in range(1, T):  # t=2 -> t=T
        a = py_x[:, ii] * (A @ norm_a[:, ii-1])
        norm_a[:, ii] = a / a.sum()
        logL[ii] = np.log(a.sum())  # one-step predictive log likelihood

    p_h_t_v1_t = norm_a  # filtered hidden state posterior density

    # Compute the beta (backward pass)
    norm_b = np.zeros((K, T))
    norm_b[:, -1] = np.ones(K)  # b(h_T) = 1 for all h_T at t=T
    for ii in range(T-2, -1, -1):  # t=T-1 -> t=1
        b = A.T @ (py_x[:, ii+1] * norm_b[:, ii+1])
        norm_b[:, ii] = b / b.sum()

    # norm_a and norm_b are defined for t=1 -> t=T
    ab = norm_a * norm_b
    p_h_t_v1_T = ab / ab.sum(axis=0)  # smoothed hidden state posterior density

    # Compute pairwise edge marginals using a and b (note that they are normalized)
    if compute_edge:
        p_h_t_tmin1_v1_T = np.zeros((K, K, T))
        for ii in range(1, T):  # t=2 -> t=T
            ab_mat = norm_a[:, ii-1] * py_x[:, ii][:, None] * A * norm_b[:, ii][:, None]
            p_h_t_tmin1_v1_T[:, :, ii] = ab_mat / ab_mat.sum()  # P(h_t-1,h_t|v_1:T)
    else:
        p_h_t_tmin1_v1_T = None

    return p_h_t_v1_t, p_h_t_v1_T, p_h_t_tmin1_v1_T, logL


def baum_welch(A_init, py_x, p1_init, maxEM_iter=100):
    """
    Classical Baum-Welch algorithm on hidden markov model parameter estimation.
    In this implementation, the observation probability (also known as
    emission distribution) is fixed and therefore not estimated. Even though
    py_x has explicit M step update equations and is estimable by a general
    Baum-welch for both discrete and Gaussian observation variable y, it often
    has known relations that determine the observation probability thus fixed

    Reference:
        Baum, L. E., Petrie, T., Soules, G., & Weiss, N. (1970). A maximization
        technique occurring in the statistical analysis of probabilistic
        functions of Markov chains. The annals of mathematical statistics, 41(1),
        164-171.

    Inputs:
    :param A_init: transition probability matrix for the first E step
    :param py_x: observation probability (fixed)
    :param p1_init: initial prior of hidden state at t=1 for the first E step
    :param maxEM_iter: maximal number of EM iterations
    """
    # Model dimensions
    K = A_init.shape[0]
    assert py_x.shape[0] == K, 'Dimension mismatch between A_init and py_x.'
    assert p1_init.shape[0] == K, 'Dimension mismatch between A_init and p1_init.'

    # Initialize the EM loop
    A = A_init  # A_ij = P(h_t = i | h_t-1 = j), i.e. rows >> t+1, columns >> t
    p1 = p1_init
    logL_list = [float('-inf')]
    logL_delta = float('inf')
    EM_iter = 0

    # EM iterations
    while logL_delta > 10**-6 and EM_iter <= maxEM_iter:
        EM_iter = EM_iter + 1

        # E step - taking expectation under the hidden state posterior using current parameters {A, p1}
        _, p_h_t_v1_T, p_h_t_tmin1_v1_T, logL = forward_backward(A=A, py_x=py_x, p1=p1, compute_edge=True)

        # M step - MLE of parameters {A, p1}
        p1 = p_h_t_v1_T[:, 0]  # updated p1 = P(h1|v_1:T)
        edges = p_h_t_tmin1_v1_T.sum(axis=2)  # sum over time points
        A = edges / edges.sum(axis=0)  # updated A = \sum_t{P(h_t-1,h_t|v_1:T)} / \sum_t{P(h_t-1|v_1:T)}

        # Check convergence using logL
        logL = logL.sum()
        logL_delta = abs(logL - logL_list[-1])
        logL_list.append(logL)

    return A, p1, EM_iter, logL_list


def viterbi(A, py_x, p1=None, ignore_numerr=False):
    """
    Classical Viterbi algorithm for discrete state HMM to solve argmax exact
    inference. Note that we use the log-probability version to cope with
    underflow numerical precision, therefore this is an instance of the
    general Max-Sum algorithm specialized to the tree graphical model of HMM

    Reference:
        Viterbi, A. (1967). Error bounds for convolutional codes and an
        asymptotically optimum decoding algorithm. IEEE transactions on
        Information Theory, 13(2), 260-269.

    Inputs:
    :param A: transition probability matrix
    :param py_x: observation probability
    :param p1: initial prior of hidden state at t=1
    :param ignore_numerr: whether to ignore warning messages of numerical errors
    """
    # Model dimensions
    K = A.shape[0]
    T = py_x.shape[1]
    assert py_x.shape[0] == K, 'Dimension mismatch between A and py_x.'

    # Initialize prior at t=1
    p1 = np.ones(K) / K if p1 is None else p1

    if ignore_numerr:
        old_settings = np.seterr(divide='ignore')  # only ignore log(0) error

    # Initialize the trellises
    Tre_p = np.zeros((K, T))  # store max probability
    Tre_p[:, 0] = np.log(p1) + np.log(py_x[:, 0])  # initial state probability
    Tre_h = np.zeros((K, T), dtype=int)  # store argmax state

    # Forward pass to fill the trellises
    for ii in range(1, T):  # t=2 -> t=T
        P = Tre_p[:, ii-1] + np.log(A) + np.log(py_x[:, ii])[:, None]
        Tre_h[:, ii] = np.argmax(P, axis=1)
        Tre_p[:, ii] = P[range(K), Tre_h[:, ii]]

    # Backward pass to identify the global argmax path
    viterbi_path = np.zeros(T, dtype=int)
    viterbi_path[-1] = np.argmax(Tre_p[:, -1])
    for ii in range(T-2, -1, -1):  # t=T-1 -> t=1
        viterbi_path[ii] = Tre_h[viterbi_path[ii+1], ii+1]

    # Output binary hidden state path
    bin_path = np.zeros((K, T), dtype=bool)
    for ii in range(K):
        bin_path[ii, viterbi_path == ii] = 1

    if ignore_numerr:
        # noinspection PyUnboundLocalVariable
        np.seterr(**old_settings)

    return viterbi_path, bin_path


def logdet(A):
    """
    Computes logdet using Schur complement (Non-torch)

    M = [A B; C D]
    det(M) = det(A) * det(D - C @ A^-1 @ B)
    """
    log_det = np.log(A[0, 0])
    for _ in range(A.shape[0] - 1):
        A = A[1:, 1:] - A[1:, 0][:, None] @ A[0, 1:][None, :] / A[0, 0]
        log_det += np.log(A[0, 0])
    return log_det


def logdet_torch(A):
    """ Computes logdet using Schur complement """
    log_det = torch.log(A[0, 0])
    for _ in range(A.shape[0] - 1):
        A = A[1:, 1:] - torch.matmul(torch.unsqueeze(A[1:, 0], -1),
                                     torch.unsqueeze(A[0, 1:], 0)) / A[0, 0]
        log_det += torch.log(A[0, 0])
    return log_det


def apply_R_weights(R, q, T, R_weights=None):
    """ Apply time-varying weights on the observation noise covariance matrix """
    for R_val in [np.diag(R), np.linalg.eigvals(R)]:
        assert (R_val > 0.).all(), 'R must be positive definite.'

    if R_weights is None:
        R = R[:, :, None] * np.ones((q, q, T))
    else:  # (index 1 corresponds to t=1, etc.)
        R = R[:, :, None] * np.reshape(np.tile(R_weights, (q ** 2, 1)), (q, q, T))
        R[np.isnan(R)] = 0.  # when there was 0 * inf = nan, set it to be 0

    return R


def kalman_filter(F, Q, mu0, S0, G, R, y):
    """
    This is the classical Kalman filter only without smoothing.
    Multivariate observation data y is supported. This implementation runs
    the kalman filtering on CPU. This function has been heavily optimized
    for matrix computations in exchange for more intense memory use
    and arcane syntax. The bare minimum of the Kalman filter is implemented
    with the forward pass only and no extra outputs. This is created to
    facilitate the use of the Kalman filter in real-time applications.

    Reference:
        Shumway, R. H., & Stoffer, D. S. (1982). An approach to time
        series smoothing and forecasting using the EM algorithm.
        Journal of time series analysis, 3(4), 253-264.

        Jazwinski, A. H. (2007). Stochastic processes and filtering
        theory. Courier Corporation.

    Inputs
    :param F: transition matrix
    :param Q: state noise covariance matrix
    :param mu0: initial state mean vector
    :param S0: initial state covariance matrix
    :param G: observation matrix
    :param R: observation noise covariance matrix
    :param y: observed data (can be multivariate)
    :return: x_t_t, P_t_t
    """
    # Vector dimensions
    _, T = y.shape
    p = mu0.shape[0]

    # Kalman filtering (forward pass)
    x_t_t = np.zeros((p, T+1))  # (index 1 corresponds to t=0, etc.)
    P_t_t = np.zeros((p, p, T+1))

    # Initialize
    x_t_t[:, 0] = np.squeeze(mu0, axis=1)  # x_0_0
    P_t_t[:, :, 0] = S0  # P_0_0

    # Recursion of forward pass
    for ii in range(1, T+1):  # t=1 -> t=T
        # One-step prediction
        x_t_tmin1 = F @ x_t_t[:, ii-1]
        P_t_tmin1 = F @ P_t_t[:, :, ii-1] @ F.T + Q

        # Update
        GP = G @ P_t_tmin1
        Sigma = GP @ G.T + R[:, :, ii-1]
        invSigma = inverse(Sigma, approach='gaussian')
        dy = y[:, ii-1] - G @ x_t_tmin1
        K_t = GP.T @ invSigma
        x_t_t[:, ii] = x_t_tmin1 + K_t @ dy
        P_t_t[:, :, ii] = P_t_tmin1 - K_t @ GP

    return x_t_t, P_t_t


def kalman(F, Q, mu0, S0, G, R, y, R_weights=None, skip_interp=True):
    """
    This is the classical Kalman filter and fixed-interval smoother.
    Multivariate observation data y is supported. This implementation runs
    the kalman filtering and smoothing on CPU. This function has been heavily
    optimized for matrix computations in exchange for more intense memory use
    and arcane syntax.

    Reference:
        Shumway, R. H., & Stoffer, D. S. (1982). An approach to time
        series smoothing and forecasting using the EM algorithm.
        Journal of time series analysis, 3(4), 253-264.

        De Jong, P., & Mackinnon, M. J. (1988). Covariances for
        smoothed estimates in state space models. Biometrika, 75(3),
        601-602.

        Jazwinski, A. H. (2007). Stochastic processes and filtering
        theory. Courier Corporation.

    Inputs
    :param F: transition matrix
    :param Q: state noise covariance matrix
    :param mu0: initial state mean vector
    :param S0: initial state covariance matrix
    :param G: observation matrix
    :param R: observation noise covariance matrix
    :param y: observed data (can be multivariate)
    :param R_weights: time-varying weights on the observation
            noise covariance (default: uniform unit weights)
    :param skip_interp: whether to skip calculating interpolated density (always True)
    :return: x_t_n, P_t_n, P_t_tmin1_n, logL, x_t_t, P_t_t, K_t, x_t_tmin1, P_t_tmin1, fy_t_interp (nan)
    """
    # Vector dimensions
    q, T = y.shape
    p = mu0.shape[0]
    approach_y = 'svd' if q >= 5 else 'gaussian'
    approach_x = 'svd' if p >= 5 else 'gaussian'

    # Expand observation noise R to all time points
    R = apply_R_weights(R, q, T, R_weights=R_weights)

    # Kalman filtering (forward pass)
    x_t_tmin1 = np.zeros((p, T+1))  # (index 1 corresponds to t=0, etc.)
    P_t_tmin1 = np.zeros((p, p, T+1))
    K_t = np.zeros((p, q))
    x_t_t = np.zeros((p, T+1))
    P_t_t = np.zeros((p, p, T+1))
    logL = np.zeros(T)  # (index 1 corresponds to t=1)
    qlog2pi = q * np.log(2 * np.pi)

    # Initialize
    x_t_t[:, 0] = np.squeeze(mu0, axis=1)  # x_0_0
    P_t_t[:, :, 0] = S0  # P_0_0

    # Recursion of forward pass
    for ii in range(1, T+1):  # t=1 -> t=T
        # One-step prediction
        x_t_tmin1[:, ii] = F @ x_t_t[:, ii-1]
        P_t_tmin1[:, :, ii] = F @ P_t_t[:, :, ii-1] @ F.T + Q

        # Update
        GP = G @ P_t_tmin1[:, :, ii]
        Sigma = GP @ G.T + R[:, :, ii-1]
        invSigma = inverse(Sigma, approach=approach_y)
        dy = y[:, ii-1] - G @ x_t_tmin1[:, ii]
        K_t = GP.T @ invSigma
        x_t_t[:, ii] = x_t_tmin1[:, ii] + K_t @ dy
        P_t_t[:, :, ii] = P_t_tmin1[:, :, ii] - K_t @ GP

        # Innovation form of the log likelihood
        logL[ii-1] = -(qlog2pi + logdet(Sigma) + dy.T @ invSigma @ dy) / 2

    # Kalman smoothing (backward pass)
    x_t_n = np.zeros((p, T+1))  # (index 1 corresponds to t=0, etc.)
    P_t_n = np.zeros((p, p, T+1))
    P_t_tmin1_n = np.zeros((p, p, T+1))  # cross-covariance between t and t-1
    fy_t_interp = float('nan')  # interpolated conditional density is not available in classical kalman filtering

    # Initialize
    x_t_n[:, -1] = x_t_t[:, -1]  # x_T_T
    P_t_n[:, :, -1] = P_t_t[:, :, -1]  # P_T_T

    # Recursion of backward pass
    for ii in range(T, 0, -1):  # t=T -> t=1
        J_t = P_t_t[:, :, ii-1] @ F.T @ inverse(P_t_tmin1[:, :, ii], approach=approach_x)
        JP = J_t @ P_t_n[:, :, ii]
        x_t_n[:, ii-1] = x_t_t[:, ii-1] + J_t @ (x_t_n[:, ii] - x_t_tmin1[:, ii])
        P_t_n[:, :, ii-1] = P_t_t[:, :, ii-1] + (JP - J_t @ P_t_tmin1[:, :, ii]) @ J_t.T
        P_t_tmin1_n[:, :, ii] = JP.T  # Cov(t,t-1) proved in De Jong & Mackinnon (1988)

    return x_t_n, P_t_n, P_t_tmin1_n, logL, x_t_t, P_t_t, K_t, x_t_tmin1, P_t_tmin1, fy_t_interp


def djkalman(F, Q, mu0, S0, G, R, y, R_weights=None, skip_interp=True):
    """
    This is the [De Jong 1989] Kalman filter and fixed-interval smoother.
    Multivariate observation data y is supported. This implementation runs
    the kalman filtering and smoothing on CPU. This function has been heavily
    optimized for matrix computations in exchange for more intense memory use
    and arcane syntax. Note the filtered estimates are skipped since not used
    during recursion. If you need filtered estimate, refer to the equations
    derived in the MATLAB human-readable version (under <archive>) and add to
    the for loop.

    Since De Jong Kalman filtering and smoothing are not defined at t=0, we
    repeat the estimates at t=1 to extend to t=0. This is useful to allow
    updating initial state and covariance estimates in the M step.

    Reference:
        De Jong, P. (1989). Smoothing and interpolation with the
        state-space model. Journal of the American Statistical
        Association, 84(408), 1085-1088.

    Inputs
    :param F: transition matrix
    :param Q: state noise covariance matrix
    :param mu0: initial state mean vector
    :param S0: initial state covariance matrix
    :param G: observation matrix
    :param R: observation noise covariance matrix
    :param y: observed data (can be multivariate)
    :param R_weights: time-varying weights on the observation
            noise covariance (default: uniform unit weights)
    :param skip_interp: whether to skip calculating interpolated density
    :return: x_t_n, P_t_n, P_t_tmin1_n, logL, x_t_t, P_t_t, K_t, x_t_tmin1, P_t_tmin1, fy_t_interp
    """
    # Vector dimensions
    q, T = y.shape
    p = mu0.shape[0]
    approach_y = 'svd' if q >= 5 else 'gaussian'

    # Expand observation noise R to all time points
    R = apply_R_weights(R, q, T, R_weights=R_weights)

    # Kalman filtering (forward pass)
    x_t_tmin1 = np.zeros((p, T+2))  # (index 1 corresponds to t=0, etc.)
    P_t_tmin1 = np.zeros((p, p, T+2))
    K_t = np.zeros((p, q, T+1))  # note that this is different from the classical Kalman gain by pre-multiplying with F
    e_t = np.zeros((q, T+1))
    invD_t = np.zeros((q, q, T+1))
    L_t = np.zeros((p, p, T+1))
    logL = np.zeros(T)  # (index 1 corresponds to t=1)
    qlog2pi = q * np.log(2 * np.pi)

    # Initialize
    x_t_t = mu0  # x_0_0, initialized only to keep return outputs consistent
    P_t_t = S0  # P_0_0, initialized only to keep return outputs consistent
    x_t_tmin1[:, 1] = F @ np.squeeze(x_t_t, axis=1)  # x_1_0 (W_0*beta in De Jong 1989)
    P_t_tmin1[:, :, 1] = F @ P_t_t @ F.T + Q  # P_1_0 (V_0 in De Jong 1989)

    # Recursion of forward pass
    for ii in range(1, T+1):  # t=1 -> t=T
        # Intermediate vectors
        e_t[:, ii] = y[:, ii-1] - G @ x_t_tmin1[:, ii]  # same as dy in classical kalman
        D_t = G @ P_t_tmin1[:, :, ii] @ G.T + R[:, :, ii-1]  # same as Sigma in classical kalman
        invD_t[:, :, ii] = inverse(D_t, approach=approach_y)
        FP = F @ P_t_tmin1[:, :, ii]
        K_t[:, :, ii] = FP @ G.T @ invD_t[:, :, ii]
        L_t[:, :, ii] = F - K_t[:, :, ii] @ G

        # One-step prediction for the next time point
        x_t_tmin1[:, ii+1] = F @ x_t_tmin1[:, ii] + K_t[:, :, ii] @ e_t[:, ii]
        P_t_tmin1[:, :, ii+1] = FP @ L_t[:, :, ii].T + Q

        # Innovation form of the log likelihood
        logL[ii-1] = -(qlog2pi + logdet(D_t) + e_t[:, ii].T @ invD_t[:, :, ii] @ e_t[:, ii]) / 2

    # remove the extra t=T+1 time point created
    x_t_tmin1 = x_t_tmin1[:, :-1]
    P_t_tmin1 = P_t_tmin1[:, :, :-1]

    # Kalman smoothing (backward pass)
    r_t = np.zeros((p, T+1))  # (index 1 corresponds to t=0, etc.)
    R_t = np.zeros((p, p, T+1))
    x_t_n = np.zeros((p, T+1))
    P_t_n = np.zeros((p, p, T+1))
    P_t_tmin1_n = np.zeros((p, p, T+1))  # cross-covariance between t and t-1
    Ip = np.eye(p)

    # Recursion of backward pass - fixed-interval smoothing
    for ii in range(T, 0, -1):  # t=T -> t=1
        # Intermediate vectors
        GD = G.T @ invD_t[:, :, ii]
        r_t[:, ii-1] = GD @ e_t[:, ii] + L_t[:, :, ii].T @ r_t[:, ii]
        R_t[:, :, ii-1] = GD @ G + L_t[:, :, ii].T @ R_t[:, :, ii] @ L_t[:, :, ii]

        # Smoothed estimates
        x_t_n[:, ii] = x_t_tmin1[:, ii] + P_t_tmin1[:, :, ii] @ r_t[:, ii-1]
        RP = R_t[:, :, ii-1] @ P_t_tmin1[:, :, ii]
        P_t_n[:, :, ii] = P_t_tmin1[:, :, ii] @ (Ip - RP)
        # Cov(t,t-1) is derived using Theorem 1 and Lemma 1, m = t, s = t+1
        P_t_tmin1_n[:, :, ii] = (Ip - RP.T) @ L_t[:, :, ii-1] @ P_t_tmin1[:, :, ii-1].T

    # Set the cross-covariance estimate at t=1. Cov(t=1,t=0) can be computed exactly using J_0.
    # But we use P_t=1_n instead to avoid inverting conditional state noise covariance.
    P_t_tmin1_n[:, :, 1] = P_t_n[:, :, 1]

    # Repeat t=1 to extend the smoothed estimates to t=0
    x_t_n[:, 0] = x_t_n[:, 1]
    P_t_n[:, :, 0] = P_t_n[:, :, 1]

    # Interpolated conditional density of y_t
    if skip_interp:
        fy_t_interp = float('nan')
    else:
        # fy_t_interp = normpdf(y_t | y_1,...,y_t-1,y_t+1,y_T)
        fy_t_interp = np.zeros(T)  # (index 1 corresponds to t=1, etc.)
        for ii in range(T):  # t=1 -> t=T
            n_t = invD_t[:, :, ii+1] @ e_t[:, ii+1] - K_t[:, :, ii+1].T @ r_t[:, ii+1]
            N_t = invD_t[:, :, ii+1] + K_t[:, :, ii+1].T @ R_t[:, :, ii+1] @ K_t[:, :, ii+1]
            # See De Jong 1989 Section 5, note that -logdet(N_t) is NOT a typo
            if np.linalg.cond(N_t) < (1 / np.finfo(float).eps):
                invN_t = inverse(N_t, approach=approach_y)
            else:
                invN_t = np.ones(N_t.shape, dtype=np.float64) * float('inf')
            fy_t_interp[ii] = np.exp(-(qlog2pi - logdet(N_t) + n_t.T @ invN_t @ n_t) / 2)

    return x_t_n, P_t_n, P_t_tmin1_n, logL, x_t_t, P_t_t, K_t, x_t_tmin1, P_t_tmin1, fy_t_interp


def djkalman_conv_torch(F, Q, mu0, S0, G, R, y, conv_steps=100, dtype=torch.float32):
    """
    A convergent version of djkalman() implemented in Pytorch

    State space model is defined as follows:
        x(t) = F*x(t-1)+eta(t)   eta ~ N(0,Q) (state or transition equation)
        y(t) = G*x(t)+eps(t)     eps ~ N(0,R) (observation or measurement equation)

    djkalman_conv_torch provides a pytorch based implementation (for gpu) that
    computes the one-step prediction and the smoothed estimate, as well as
    their covariance matrices. The function uses forward and backward
    recursions, and uses a convergent approach to compute steady state version
    of the Kalman gain (and hence the covariance matrices) for reducing
    runtime.

    Authors: Ran Liu <rliu20@mgh.harvard.edu>
             Mingjian He <mh1@stanford.edu>

    Input:
    -----
    F: Nx x Nx matrix
        time-invariant transition matrix in transition equation.
    Q: Nx x Nx matrix
        time-invariant covariance matrix for the error in transition equation.
    mu0: Nx x 1
        initial state mean vector.
    S0: Nx x Nx
        covariance matrix of the initial state vector.
    G: Ny x Nx matrix
        time-invariant measurement matrix in measurement equation.
    R: Ny x Ny matrix
        time-invariant covariance matrix for the error in measurement equation.
    y: Ny x T matrix
        containing data (y(1), ... , y(T)).
    conv_steps: int (default 100)
        Kalman gain is updated up to this many steps
    dtype: cuda tensor data type (default torch.float32)
        the tensor precision data type for GPU processing

    Output:
    ------
    x_t_n: Nx x T matrix
        smoothed state vectors.
    P_t_n: Nx x Nx matrix
        SS covariance matrices of smoothed state vectors.
    P_t_tmin1_n: Nx x Nx matrix
        SS cross-covariance (lag 1) matrices of smoothed state vectors.
    logL: 1 x T vector (float)
        value of the log likelihood function of the SSM at each time point
        under assumption that observation noise eps(t) is normally distributed.
    break_conv: logical flag (boolean)
        whether numerical accuracy limit reached during convergence
    K_t: Nx x Nx matrix
        SS Kalman gain.
    x_t_tmin1: Nx x T matrix
        one-step predicted state vectors.
    P_t_tmin1: Nx x Nx matrix
        SS mean square error of predicted state vectors.
    """
    # Vector dimensions
    q, T = y.shape
    p = mu0.shape[0]

    # Kalman filtering (forward pass)
    x_t_tmin1 = torch.zeros(p, T+2, dtype=dtype).cuda()  # (index 1 corresponds to t=0, etc.)
    K_t = torch.zeros(p, q, dtype=dtype).cuda()
    e_t = torch.zeros(q, T+1, dtype=dtype).cuda()
    # noinspection PyUnusedLocal
    D_t = torch.zeros(q, q, dtype=dtype).cuda()
    invD_t = torch.zeros(q, q, dtype=dtype).cuda()
    log_det_Dt = torch.zeros(1).cuda()
    L_t = torch.zeros(p, p, dtype=dtype).cuda()
    logL = torch.zeros(T, dtype=dtype).cuda()  # (index 1 corresponds to t=1)
    qlog2pi = q * torch.log(torch.as_tensor(2 * torch.pi, dtype=dtype).cuda())
    break_conv = False

    # Initialize
    x_t_tmin1[:, 1] = torch.matmul(F, mu0)[:, 0]  # x_1_0 (W_0*beta in De Jong 1989)
    P_t_tmin1 = torch.matmul(torch.matmul(F, S0), F.T) + Q  # P_1_0 (V_0 in De Jong 1989)

    # Recursion of forward pass until convergence of predicted covariance matrix
    for ii in range(1, conv_steps+1):  # t=1 -> t=conv_steps
        # Intermediate vectors
        e_t[:, ii] = y[:, ii-1] - torch.matmul(G, x_t_tmin1[:, ii])
        FP = torch.matmul(F, P_t_tmin1)

        # Check for numerical precision limit
        temp_D_t = torch.matmul(torch.matmul(G, P_t_tmin1), G.T) + R
        temp_invD_t = inverse_torch(temp_D_t, approach='svd')
        temp_K_t = torch.matmul(torch.matmul(FP, G.T), temp_invD_t)
        temp_L_t = F - torch.matmul(temp_K_t, G)
        temp_P_t_tmin1 = torch.matmul(FP, temp_L_t.T) + Q

        # If precision limit reached, must accept last variables as converged
        if torch.as_tensor(torch.diag(temp_P_t_tmin1) < 0).any():
            break_conv = True
            invD_t = invD_t * float('nan') if ii < 10 else invD_t  # set invD_t to nan if too few iterations
        else:
            D_t = temp_D_t
            invD_t = temp_invD_t
            K_t = temp_K_t
            L_t = temp_L_t
            P_t_tmin1 = temp_P_t_tmin1

        # One-step prediction for the next time point
        x_t_tmin1[:, ii+1] = torch.matmul(F, x_t_tmin1[:, ii]) + torch.matmul(K_t, e_t[:, ii])

        # Innovation form of the log likelihood
        log_det_Dt = logdet_torch(D_t)  # logdet of prediction error covariance also converges
        logL[ii-1] = -(qlog2pi + log_det_Dt + torch.matmul(
            torch.matmul(torch.unsqueeze(e_t[:, ii], 0), invD_t),
            torch.unsqueeze(e_t[:, ii], -1))) / 2

        if break_conv:
            conv_steps = ii
            break

    # Recursion of forward pass for the remaining time steps
    for ii in range(conv_steps+1, T+1):  # t=conv_steps+1 -> t=T
        e_t[:, ii] = y[:, ii-1] - torch.matmul(G, x_t_tmin1[:, ii])
        x_t_tmin1[:, ii+1] = torch.matmul(F, x_t_tmin1[:, ii]) + torch.matmul(K_t, e_t[:, ii])
        logL[ii-1] = -(qlog2pi + log_det_Dt + torch.matmul(
            torch.matmul(torch.unsqueeze(e_t[:, ii], 0), invD_t),
            torch.unsqueeze(e_t[:, ii], -1))) / 2

    x_t_tmin1 = x_t_tmin1[:, :-1]  # Remove the extra t=T+1 time point created

    # Kalman smoothing (backward pass)
    r_t = torch.zeros(p, dtype=dtype).cuda()
    x_t_n = torch.zeros(p, T+1, dtype=dtype).cuda()  # (index 1 corresponds to t=0, etc.)
    Ip = torch.eye(p, dtype=dtype).cuda()

    # Run R_t recursion until convergence
    GD = torch.matmul(G.T, invD_t)
    R_t = torch.zeros(p, p, dtype=dtype).cuda()  # dlyap is slow on GPU, find convergent R_t empirically
    GDG = torch.matmul(GD, G)

    for k in range(conv_steps):  # t=T -> t=T-conv_steps+1
        R_t = GDG + torch.matmul(torch.matmul(L_t.T, R_t), L_t)

    RP = torch.matmul(R_t, P_t_tmin1)
    P_t_n = torch.matmul(P_t_tmin1, (Ip - RP))
    P_t_tmin1_n = torch.matmul(
        torch.matmul((Ip - RP.T), L_t), P_t_tmin1.T)  # derived using Theorem 1 and Lemma 1, m = t, s = t+1

    # Recursion of backward pass: fixed-interval smoothing
    for ii in range(T, 0, -1):  # t=T -> t=1
        r_t = torch.matmul(GD, e_t[:, ii]) + torch.matmul(L_t.T, r_t)
        x_t_n[:, ii] = x_t_tmin1[:, ii] + torch.matmul(P_t_tmin1, r_t)

    x_t_n[:, 0] = x_t_n[:, 1]  # Repeat t=1 to extend smoothed estimates to t=0

    return x_t_n, P_t_n, P_t_tmin1_n, logL, break_conv, K_t, x_t_tmin1, P_t_tmin1


def inverse(A, approach='svd'):
    """
    Custom inverse function for inverting large covariance matrices

    Several approaches are possible:
        - 'gaussian': Gaussian elimination
        - 'cholesky': Cholesky decomposition
        - 'qr': QR decomposition
        - 'svd': Singular value decomposition

    Authors: Proloy Das <pd640@mgh.harvard.edu>
             Mingjian He <mh1@stanford.edu>
    """
    if np.isinf(A).any():
        return np.zeros(A.shape, dtype=A.dtype)

    if approach == 'gaussian':
        return np.linalg.inv(A)

    if approach == 'cholesky':
        L = np.linalg.cholesky(A)
        L_inv = solve_triangular(L, np.eye(L.shape[0], dtype=L.dtype), lower=True)
        return L_inv.T @ L_inv

    elif approach == 'qr':
        Q, R = np.linalg.qr(A, mode='complete')
        R_inv = solve_triangular(R, np.eye(R.shape[0], dtype=R.dtype), lower=False)
        return R_inv @ Q.T

    elif approach == 'svd':
        U, S, Vh = np.linalg.svd(A, full_matrices=True)
        S = _reciprocal_pos_vals(S)
        return (Vh.T * S) @ U.T

    else:
        raise ValueError('Specified matrix inversion approach is not recognized.')


def inverse_torch(A, approach='svd'):
    """
    Custom inverse function for inverting large covariance matrices with Pytorch

    see function header of inverse() for different approaches
    """
    if approach == 'gaussian':  # not recommended
        return torch.linalg.inv(A)

    elif approach == 'cholesky':  # if on CUDA will use CPU therefore slow
        L = torch.linalg.cholesky(A, upper=False)
        Ia = torch.eye(L.shape[0], dtype=L.dtype, device=L.device)
        L_inv = torch.linalg.solve_triangular(L, Ia, upper=False)
        return torch.matmul(L_inv.T, L_inv)

    elif approach == 'qr':
        Q, R = torch.linalg.qr(A, mode='complete')
        Ir = torch.eye(R.shape[0], dtype=R.dtype, device=R.device)
        R_inv = torch.linalg.solve_triangular(R, Ir, upper=True)
        return torch.matmul(R_inv, Q.T)

    elif approach == 'svd':  # if on CUDA will use GPU unless A is complex
        U, S, Vh = torch.linalg.svd(A, full_matrices=True)
        S = _reciprocal_pos_vals(S)
        return torch.matmul(Vh.T * S, U.T)

    else:
        raise ValueError('Specified matrix inversion approach is not recognized.')


def _reciprocal_pos_vals(S, epsilon=1e-5):
    """ Take the reciprocal of positive values and set the rest to zero """
    S_pos_idx = (S / S[0]) > epsilon  # find indices of positive (large enough) singular values
    S[~S_pos_idx] = 0.0  # set the remaining singular values to 0
    S[S_pos_idx] = 1.0 / S[S_pos_idx]  # reciprocal of positive singular values
    return S



"""
Author: Mingjian He <mh1@stanford.edu>

Testing functions for switching methods in somata/switching
"""

import numpy as np
from test_load_data import _load_data  # type: ignore


def test_switching():
    from somata.switching import switching
    from somata import OscillatorModel as Osc

    input_data = _load_data("switching_input.mat")

    o1 = Osc(**input_data, G=([1, 0, 1, 0], [1, 0, 0, 0]), Fs=100)
    print(o1)

    method_list = ('static', 'gpb1', 'gpb2', 'imm', '1991',
                   'a parp', 'ab parp', 'a pari', 'ab pari', 'a pars', 'ab pars')
    for method in method_list:
        Mprob, fy_t = switching(o1, method=method)
        mat_results = _load_data(method + '_results.mat')

        # Verify that the results are close within tolerance
        assert np.allclose(Mprob.flatten(), mat_results['Mprob'].flatten()), method + 'Mprob results are off.'
        assert np.allclose(fy_t.flatten(), mat_results['fy_t'].flatten()), method + 'fy_t results are off.'

    # Test 1991 with 5 future steps of pseudo-smoothing
    Mprob, fy_t = switching(o1, method='1991', future_steps=5)
    mat_results = _load_data('1991_future5_results.mat')
    assert np.allclose(Mprob.flatten(), mat_results['Mprob'].flatten()), '1991 ps=5 Mprob results are off.'
    assert np.allclose(fy_t.flatten(), mat_results['fy_t'].flatten()), '1991 ps=5 fy_t results are off.'


def test_vb_learn_original():
    from somata.switching import VBSwitchModel as Vbs
    from somata import OscillatorModel as Osc

    input_data = _load_data("switching_input.mat")

    o1 = Osc(**input_data, G=([1, 0, 1, 0], [1, 0, 0, 0]), Fs=100)
    print(o1)

    # SOMATA has better estimations of mu0 and S0, so remove them from EM updates when comparing with MATLAB results
    vb_model = Vbs(o1)
    print(vb_model)
    vb_results = vb_model.learn(keep_param=('mu0', 'S0'), return_dict=True, original=True)
    mat_results = _load_data("vb_learn_original_results.mat")

    # Verify that the results are close within tolerance
    assert np.allclose(vb_results['h_t_m'].flatten(), mat_results['Mprob'].flatten()), \
        'Results of h_t_m are different from MATLAB outputs.'
    assert np.allclose(vb_results['q_t_m'].flatten(), mat_results['fy_t'].flatten()), \
        'Results of q_t_m are different from MATLAB outputs.'
    assert np.allclose(vb_results['A'].flatten(), mat_results['A'].flatten()), \
        'Results of A are different from MATLAB outputs.'
    assert np.allclose(vb_results['VB_iter'], mat_results['VB_iter'].flatten()), \
        'Different numbers of VB iterations were executed from MATLAB outputs.'
    assert np.allclose([x[-1] for x in vb_results['logL_bound']], mat_results['logL_bound'].flatten()), \
        'Results of logL_bound are different from MATLAB outputs.'


def test_vb_learn():
    from somata.switching import VBSwitchModel as Vbs
    from somata import OscillatorModel as Osc

    input_data = _load_data("switching_input.mat")

    o1 = Osc(**input_data, G=([1, 0, 1, 0], [1, 0, 0, 0]), Fs=100)
    print(o1)

    # SOMATA has better estimations of mu0 and S0, so remove them from EM updates when comparing with MATLAB results
    vb_model = Vbs(o1)
    print(vb_model)
    vb_results = vb_model.learn(keep_param=('mu0', 'S0'), shared_R=True, shared_comp=[0, 0],
                                normalize_q_t=True, return_dict=True)
    mat_results = _load_data("vb_learn_results.mat")

    # Verify that the results are close within tolerance
    assert np.allclose(vb_results['h_t_m'].flatten(), mat_results['Mprob'].flatten()), \
        'Results of h_t_m are different from MATLAB outputs.'
    assert np.allclose(vb_results['h_t_m_soft'].flatten(), mat_results['Mprob_soft'].flatten()), \
        'Results of h_t_m_soft are different from MATLAB outputs.'
    assert np.allclose(vb_results['h_t_m_hard'].flatten(), mat_results['Mprob_hard'].flatten()), \
        'Results of h_t_m_hard are different from MATLAB outputs.'
    assert np.allclose(vb_results['q_t_m'].flatten(), mat_results['fy_t'].flatten()), \
        'Results of q_t_m are different from MATLAB outputs.'
    assert np.allclose(vb_results['A'].flatten(), mat_results['A'].flatten()), \
        'Results of A are different from MATLAB outputs.'
    assert np.allclose(vb_results['VB_iter'], mat_results['VB_iter'].flatten()), \
        'Different numbers of VB iterations were executed from MATLAB outputs.'
    assert np.allclose([x[-1] for x in vb_results['logL_bound']], mat_results['logL_bound'].flatten()), \
        'Results of logL_bound are different from MATLAB outputs.'


if __name__ == "__main__":
    test_switching()
    test_vb_learn_original()
    test_vb_learn()
    print('Switching inference tests finished without exception.')



"""
Author: Ran Liu <rliu20@stanford.edu>

Implementation of phase-amplitude coupling estimation methods
"""

import numpy as np
import pathlib
from importlib.util import find_spec
from somata.basic_models import StateSpaceModel
from somata.exact_inference import inverse
from numpy import ndarray
from cmdstanpy import CmdStanModel
from scipy.optimize import minimize, Bounds
from typing import List
import logging


def pac_anchor_func():
    """Anchor function for locating this .py file"""
    return


def get_pac_path():
    """Find the path to the directory containing the pac_model.py file"""
    f_path = pathlib.Path(pac_anchor_func.__code__.co_filename)
    if f_path.is_file():
        return f_path.parents[0].resolve()
    else:
        somata_path = pathlib.Path(
            find_spec("somata").origin
        )  # find the package root __init__.py filepath
        return (somata_path.parents[0] / "pac").resolve()  # 1-level up


def fit_pac_regression(phase: ndarray, amplitude: ndarray, verbose=True, **kwargs):
    """
    Fit a constrained regression between phase and amplitude according to Soulat et al. 2022.
    Note that this implementation fixes an error in the manuscript that incorrectly
    uses a truncated multivariate t-distribution as the likelihood for numerical search.

    :param phase: Numpy array containing phase vector, of shape (n,).
    :param amplitude: Numpy array containing amplitude vector of shape (n,).
    :param verbose: Whether to print CmdStanPy output to stdout.
    :param **kwargs: Additional parameters to pass to model.sample() for Stan model fitting.

    :return: Draws from posterior distribution of parameters (beta0, beta1, beta2), given observed data.
    """
    model_data = {"n": len(phase), "phase": phase, "amplitude": amplitude}
    model = CmdStanModel(stan_file=get_pac_path() / "stan/pac.stan")

    if not verbose:
        logger = logging.getLogger('cmdstanpy')
        cmdstanpy_logger_status = logger.disabled
        logger.disabled = True

    fit = model.sample(model_data, show_progress=verbose, **kwargs)

    if not verbose:
        logger.disabled = cmdstanpy_logger_status

    # Empirically, these draws seem to follow a multivariate t-distribution;
    # therefore, posterior mean is a reasonable estimator of posterior mode.
    return fit.draws_pd()[["beta0", "beta1", "beta2"]].to_numpy()


def kmod(beta0, beta1, beta2):
    """
    Computes modulation index, given fitted coefficients, according to Soulat et al. 2022.

    :param beta0: Regression intercept.
    :param beta1: Regression coefficient for cos(phase).
    :param beta2: Regression coefficient for sin(phase).

    :return: Computed modulation index.
    """
    return np.sqrt(beta1 ** 2 + beta2 ** 2) / beta0


def phimod(beta1, beta2):
    """
    Computes mean modulation phase, given fitted coefficients, according to Soulat et al. 2022.

    :param beta1: Regression coefficient for cos(phase).
    :param beta2: Regression coefficient for sin(phase).

    :return: Computed mean modulation phase.
    """
    return np.arctan2(beta2, beta1)


def autocovariances(y: ndarray, p: int, rowvar=True, **kwargs):
    """
    Computes autocovariance matrices for lags from 0 to p (inclusive).

    :param y: An ndarray of n observations of d variables. If rowvars is true, expects shape (d, n); otherwise (n, d).
    :param p: Maximum lag for which to compute autocovariance.
    :param rowvar: Whether observations or variables are in first dimension. Defaults to true.
    :param **kwargs: Additional arguments to pass to numpy.cov.

    :return: A list of length p + 1 of autocovariance matrices of shape (d, d), where index i corresponds to a lag of i.
    """
    if rowvar:
        q, n = y.shape
        return [
            np.cov(y[:, :(n - i)], y[:, i:n], **kwargs)[:q, q:] for i in range(p + 1)
        ]
    else:
        n, q = y.shape
        return [
            np.cov(y[:(n - i), :], y[i:n, :], rowvar=False, **kwargs)[:q, q:]
            for i in range(p + 1)
        ]


def block_toeplitz(acov: List[ndarray]):
    """
    Constructs block-toeplitz matrix from autocovariance matrices.

    :param acov: A list of autocovariance matrices of shape (d,d). Index i corresponds to a lag of i.

    :return: A block-toeplitz matrix of shape (nxd, nxd), where the i,jth dxd block is C_{|i-j|}, and
    upper triangular blocks are transposed.
    """
    k = len(acov)  # maximum lag + 1
    q = acov[0].shape[0]
    result = np.zeros((k * q, k * q), dtype=np.float64)

    for i in range(k):
        for j in range(i + 1):  # fill diagonal and lower triangular blocks
            result[(i * q):((i + 1) * q), (j * q):((j + 1) * q)] = acov[i-j]
        for j in range(i + 1, k):  # transpose upper triangular blocks
            result[(i * q):((i + 1) * q), (j * q):((j + 1) * q)] = acov[j-i].T

    return result


def ar_parameters(acov: List[ndarray], R: float):
    """
    Compute AR parameters from autocovariance sequence and candidate observation noise variance.

    :param acov: A list of length p + 1 of autocovariance matrices of shape (d,d). Index i corresponds to a lag of i.
    :param R: Candidate observation noise variance, assumed to be the same aross all dimensions.

    :return: A tuple of a list of length p AR parameter matrices {A_i} such that
    X_t = \\sum_{k=1}^{p} A_k @ X_{t-k} + U_t, U_t ~ N(0, Q), and the state covariance matrix Q.
    """
    p = len(acov) - 1
    q = acov[0].shape[0]

    C_yule = block_toeplitz(acov)[:-q, :-q] - R * np.eye(p * q, dtype=np.float64)
    A_t = inverse(C_yule, approach='svd') @ np.concatenate(acov[1:], axis=0)
    # Compute the transpose of each block
    A = [A_t[(i * q):((i + 1) * q), :].T for i in range(p)]

    # Compute Q from C_0 = \\sum_{k=1}^{p} A_k @ C_k + Q + R (R here is a diagonal matrix)
    Q = (
        acov[0]  # C_0: autocovariance matrix at lag 0
        - R * np.eye(q, dtype=np.float64)  # diagonal observation noise covariance matrix
        - np.sum(
            np.concatenate(
                [np.expand_dims(A[i] @ acov[i + 1], axis=-1) for i in range(p)],
                axis=-1,
            ),
            axis=-1,
        )  # \\sum_{k=1}^{p} A_k @ C_k
    )

    return A, Q


def mvar_ssm(y: ndarray, A: List[ndarray], Q: ndarray, R: float):
    """
    Initializes StateSpaceModel using multivariate AR(p) parameters.

    :param y: Array of n observations of d variables, of shape (d, n).
    :param A: List of AR coefficients of length p, each element an ndarray of shape (d, d).
    :param Q: State noise covariance matrix of shape (d, d).
    :param R: Candidate observation noise variance, assumed to be the same aross all dimensions.

    :return: StateSpaceModel equivalent to specified multivariate AR model.
    """

    p = len(A)  # Order of AR model
    q = A[0].shape[0]  # Dimensionality of observations

    # Compute F (state transition matrix) from matrix coefficients in the list A
    F = np.zeros((p * q, p * q), dtype=np.float64)
    F[:q, :] = np.concatenate(A, axis=1)
    for i in range(1, p):
        F[(i * q):((i + 1) * q), ((i - 1) * q):(i * q)] = np.eye(q, dtype=np.float64)

    # Augment Q to time-lagged hidden states
    Q_new = np.zeros((p * q, p * q), dtype=np.float64)
    Q_new[:q, :q] = Q

    # Compute mu0
    mu0 = np.zeros(p * q, dtype=np.float64)

    # Compute S0 using the diagonal of Q
    S0 = np.diag(np.tile(np.diag(Q), p))

    # Compute G (observation matrix on augmented hidden states)
    G = np.zeros((q, p * q), dtype=np.float64)
    G[:q, :q] = np.eye(q, dtype=np.float64)

    # Convert R
    R_new = R * np.eye(q, dtype=np.float64)

    return StateSpaceModel(F=F, Q=Q_new, mu0=mu0, S0=S0, G=G, R=R_new, y=y)


def optimize_arp(y: ndarray, p: int, epsilon: float = 1e-7):
    """
    For a set of observations, compute the optimal observation noise variance and resulting AR(p) parameters.

    :param y: Array of n observations of d varibles, of shape (d, n).
    :param p: Order of AR model.
    :param epsilon: Epsilon to prevent singular matrices, as scipy optimizes over closed intervals, defaults to 1e-7.

    :return: AR coefficients A, state noise covariance Q, and optimal observation noise variance R.
    """
    acov = autocovariances(y, p=p, bias=True)
    C = block_toeplitz(acov)

    upper = np.min(np.linalg.eigvals(C))

    def objective_function(R):
        A, Q = ar_parameters(acov, R=R)
        model = mvar_ssm(y, A, Q, R)
        return -model.kalman_filt_smooth(return_dict=True)["logL"].sum()

    R = minimize(
        objective_function, x0=upper / 2, bounds=Bounds(lb=epsilon, ub=upper - epsilon)
        ).x[0]

    A, Q = ar_parameters(acov, R=R)

    return A, Q, R



"""
Author: Mingjian He <mh1@stanford.edu>
        Amanda Beck <ambeck@mit.edu>

osc module contains Matsuda oscillator specific methods used in SOMATA
"""

from somata.basic_models import StateSpaceModel as Ssm
import numpy as np
import numbers
from math import atan2, sin, cos, sqrt
from scipy.linalg import block_diag


class OscillatorModel(Ssm):
    """
    OscillatorModel is a subclass of StateSpaceModel class
    dedicated to Matsuda oscillators
    """
    type = 'osc'
    default_G = np.array([[1, 0]], dtype=np.float64)  # (Real, Imaginary)
    default_a = 0.99
    default_sigma2 = 3.
    a = None
    freq = None
    w = None
    sigma2 = None
    dc_idx = None
    whitener = None
    y_whiten = None

    def __init__(self, a=None, freq=None, w=None, sigma2=None, add_dc=False,
                 components='Osc', F=None, Q=None, mu0=None, S0=None, G=None, R=None, y=None, Fs=None):
        """
        Constructor method for OscillatorModel class
        :param a: damping factor of oscillators
        :param freq: rotation frequency of oscillators in hertz
        :param w: rotation frequency of oscillators in radian
        :param sigma2: noise variance of oscillators
        :param add_dc: flag to add a DC oscillator
        :param components: a list of independent oscillator components
        :param F: transition matrix
        :param Q: state noise covariance matrix
        :param mu0: initial state mean vector
        :param S0: initial state covariance matrix
        :param G: observation matrix (row major)
        :param R: observation noise covariance matrix
        :param y: observed data (row major, can be multivariate)
        :param Fs: sampling frequency in Hz
        """
        # Verify that components are of valid type
        assert components == 'Osc' or \
               all([x.type == 'osc' for x in components]), 'Encountered non-osc type component.'  # type: ignore

        # Oscillator models can be constructed by directly specifying
        # state equation oscillator parameters {a, w, sigma2}
        if freq is not None:  # one can provide frequency in Hz as well
            self.freq = np.asanyarray([freq], dtype=np.float64) if \
                isinstance(freq, numbers.Number) else np.asanyarray(freq, dtype=np.float64)
            assert Fs is not None, 'Input frequency in Hz but no sampling frequency provided.'
            w_temp = [OscillatorModel.hz_to_rad(x, Fs) for x in self.freq] if \
                self.freq.size > 1 else OscillatorModel.hz_to_rad(self.freq, Fs)
            if w is not None:
                assert (w_temp == w).all(), 'Input frequency parameters do not agree with each other.'
            w = w_temp

        if w is not None:
            w = np.asanyarray([w], dtype=np.float64) if \
                isinstance(w, numbers.Number) else np.asanyarray(w, dtype=np.float64)
            if a is not None:
                a = np.asanyarray([a], dtype=np.float64) if \
                    isinstance(a, numbers.Number) else np.asanyarray(a, dtype=np.float64)
                assert a.size == w.size, 'Different numbers of oscillator parameters provided.'
            else:
                a = np.ones_like(w, dtype=np.float64) * OscillatorModel.default_a
            if sigma2 is not None:
                sigma2 = np.asanyarray([sigma2], dtype=np.float64) if \
                    isinstance(sigma2, numbers.Number) else np.asanyarray(sigma2, dtype=np.float64)
                assert sigma2.size == w.size, 'Different numbers of oscillator parameters provided.'
            else:
                sigma2 = np.ones_like(w, dtype=np.float64) * OscillatorModel.default_sigma2

            self.a = a
            self.w = w
            self.sigma2 = sigma2

            F_tmp, Q_tmp = self._osc_to_ssm_param()
            if F is not None:
                assert (F == F_tmp).all(), 'Input state equation parameters do not agree with input F.'  # type: ignore
            if Q is not None:
                assert (Q == Q_tmp).all(), 'Input state equation parameters do not agree with input Q.'  # type: ignore
            F, Q = F_tmp, Q_tmp
        else:
            assert a is None, 'No frequency provided but damping factor a is given as input.'
            assert sigma2 is None, 'No frequency provided but noise variance sigma2 is given as input.'

        # Provide default values for mu0 and S0
        mu0 = np.zeros((F.shape[1], 1), dtype=np.float64) if mu0 is None and F is not None else mu0
        S0 = Q if S0 is None and Q is not None else S0

        # Call parent class constructor
        super().__init__(components=components, F=F, Q=Q, mu0=mu0, S0=S0, G=G, R=R, y=y, Fs=Fs)

        # Fill oscillator parameters
        self.fill_osc_param()

        # Add DC oscillator if specified
        if add_dc:
            if self.w is None:  # shortcut to create a DC oscillator using OscillatorModel(add_dc=True)
                o_dc = OscillatorModel(a=a, w=0., sigma2=sigma2, add_dc=False,
                                       F=F, Q=Q, mu0=mu0, S0=S0, R=R, y=y, Fs=Fs)
                attr_dict = o_dc.__dict__
                for attr in attr_dict.keys():
                    setattr(self, attr, getattr(o_dc, attr))
                self.dc_idx = 0
            else:
                self.add_dc()
        else:
            self.dc_idx = None

    # Dunder methods - magic methods
    def __repr__(self):
        """ Unambiguous and concise representation when calling OscillatorModel() """
        return super().__repr__().replace('Ssm', 'Osc')

    def __str__(self):
        """ Helpful information when calling print(OscillatorModel()) """
        print_str = super().__str__().replace('<Ssm object at', '<Osc object at')
        # Append additional information about oscillator parameters
        precision_backup = np.get_printoptions()['precision']
        np.set_printoptions(precision=3)
        print_str += "{0:9} = {1}\n ".format("damping a", str(self.a))
        if self.Fs is None:
            print_str += "{0:9} = {1}\n ".format("freq w", str(self.w))
        else:
            np.set_printoptions(precision=2)
            print_str += "{0:9} = {1}\n ".format("freq Hz", str(self.freq))
            np.set_printoptions(precision=3)
        print_str += "{0:9} = {1}\n ".format("sigma2", str(self.sigma2))
        print_str += "{0:9} = {1}\n ".format("obs noise R", str(self.R))
        print_str += "{0:9} = {1}\n".format("dc index", str(self.dc_idx))
        np.set_printoptions(precision=precision_backup)
        return print_str

    # Syntactic sugar methods - useful methods to make manipulations easier
    def concat_(self, other, skip_components=False):
        """
        Join two OscillatorModel objects together by concatenating the
        components.
        """
        assert self.type == 'osc', 'self in concat_() needs to be of OscillatorModel class.'

        if other.type == self.type:  # concatenation within OscillatorModel class
            # Fill oscillator parameters in both objects first
            self.update_comp_param()
            other.update_comp_param()

            # Handle DC oscillators if present
            save_other_dc = None
            save_self_dc_index = None
            if other.dc_idx is not None:
                if self.dc_idx is None:
                    save_other_dc = other.components[other.dc_idx]
                else:
                    save_self_dc_index = self.dc_idx
                # Confirm that the DC oscillator has default values therefore can be added back with constructor
                assert other.a[other.dc_idx] == other.default_a, \
                    'DC oscillator damping factor will be lost by concat_().'
                assert other.w[other.dc_idx] == 0, \
                    'DC oscillator frequency will be lost by concat_().'
                assert other.sigma2[other.dc_idx] == other.default_sigma2, \
                    'DC oscillator sigma2 will be lost by concat_().'
                # Temporarily remove the DC oscillator from the other object
                tmp_other = other.copy()
                tmp_other.remove_component(tmp_other.dc_idx)
                tmp_other.components = other.components[1:] if tmp_other.ncomp > 0 else tmp_other.components
                other = tmp_other

            # Frequency is a mandatory input for objects to call concat_()
            assert self.w is not None and other.w is not None, \
                'Both objects need at least w specified in order to concat.'
            w = np.hstack([self.w, other.w])

            # Configure the rest of attributes that are immutable
            # a
            if self.a is None and other.a is None:
                a = None
            elif self.a is None:
                a = np.hstack([np.ones_like(self.w, dtype=np.float64) * OscillatorModel.default_a, other.a])
            elif other.a is None:
                a = np.hstack([self.a, np.ones_like(other.w, dtype=np.float64) * OscillatorModel.default_a])
            else:
                a = np.hstack([self.a, other.a])

            # sigma2
            if self.sigma2 is None and other.sigma2 is None:
                sigma2 = None
            elif self.sigma2 is None:
                sigma2 = np.hstack([np.ones_like(self.w, dtype=np.float64) * OscillatorModel.default_sigma2,
                                    other.sigma2])
            elif other.sigma2 is None:
                sigma2 = np.hstack([self.sigma2,
                                    np.ones_like(other.w, dtype=np.float64) * OscillatorModel.default_sigma2])
            else:
                sigma2 = np.hstack([self.sigma2, other.sigma2])

            # S0
            if self.S0 is None and other.S0 is None:
                S0 = None
            elif self.S0 is None:
                S0 = block_diag(self.Q, other.S0) if self.Q is not None else \
                    block_diag(np.eye(self.w.size * 2, dtype=np.float64) * OscillatorModel.default_sigma2, other.S0)
            elif other.mu0 is None:
                S0 = block_diag(self.S0, other.Q) if other.Q is not None else \
                    block_diag(self.S0, np.eye(other.w.size * 2, dtype=np.float64) * OscillatorModel.default_sigma2)
            else:
                S0 = block_diag(self.S0, other.S0)

            # Call parent class method to obtain general concatenated attributes
            temp_obj = super().concat_(other, skip_components=skip_components)
            add_dc = save_other_dc is not None
            new_obj = OscillatorModel(a=a, w=w, sigma2=sigma2, add_dc=add_dc, components=temp_obj.components,
                                      mu0=temp_obj.mu0, S0=S0,
                                      R=temp_obj.R, y=temp_obj.y, Fs=temp_obj.Fs)  # fill G automatically
            new_obj.components[0] = save_other_dc if add_dc else new_obj.components[0]
            new_obj.dc_idx = save_self_dc_index if save_self_dc_index is not None else new_obj.dc_idx

        else:  # concatenation with other classes of SOMATA basic models
            new_obj = super().concat_(other, skip_components=skip_components)

        return new_obj

    def remove_component(self, comp_idx=0):
        """
        Remove a component from the OscillatorModel object,
        default to remove the left most component
        """
        if isinstance(comp_idx, list):  # recursive case
            comp_idx_list = list(np.sort(comp_idx)[::-1])
            for comp_idx in comp_idx_list:
                self.remove_component(comp_idx=comp_idx)
        else:  # base case
            super().remove_component(comp_idx=comp_idx)
            if self.dc_idx is not None:
                if comp_idx == self.dc_idx:
                    self.dc_idx = None
                elif comp_idx < self.dc_idx:
                    self.dc_idx -= 1
            if self.a is not None:
                self.a = np.delete(self.a, comp_idx)
            if self.w is not None:
                self.w = np.delete(self.w, comp_idx)
            if self.freq is not None:
                self.freq = np.delete(self.freq, comp_idx)
            if self.sigma2 is not None:
                self.sigma2 = np.delete(self.sigma2, comp_idx)

            # Set attributes to default values if nothing left
            self.a = None if len(self.a) == 0 else self.a
            self.w = None if len(self.w) == 0 else self.w
            self.freq = None if len(self.freq) == 0 else self.freq
            self.sigma2 = None if len(self.sigma2) == 0 else self.sigma2

    def add_dc(self):
        """ Add a DC oscillator to an existing OscillatorModel object """
        assert self.dc_idx is None, 'A DC oscillator already exists.'
        # usually DC oscillator has large variance therefore using max(self.sigma2)
        sigma2 = OscillatorModel.default_sigma2 if self.sigma2 is None else max(self.sigma2)
        self.rappend(OscillatorModel(w=0., sigma2=sigma2, add_dc=False))
        self.dc_idx = 0  # update the index for DC oscillator

    def fill_components(self, empty_comp=None, deep_copy=True):
        """ Fill the components attribute with OscillatorModel parameters """
        empty_comp = empty_comp or OscillatorModel()
        return super().fill_components(empty_comp=empty_comp, deep_copy=deep_copy)

    def fill_osc_param(self, F=None, Q=None, Fs=None):
        """ Attempt to fill oscillator parameters """
        F = self.F if F is None else F
        Q = self.Q if Q is None else Q
        Fs = self.Fs if Fs is None else Fs
        a, w, sigma2 = self._ssm_to_osc_param(F=F, Q=Q)
        self.a = a if a is not None else self.a
        self.w = w if w is not None else self.w
        self.sigma2 = sigma2 if sigma2 is not None else self.sigma2
        if self.w is not None and Fs is not None:
            self.freq = OscillatorModel.rad_to_hz(self.w, Fs)

    def fill_ssm_param(self, a=None, w=None, freq=None, sigma2=None, Fs=None):
        """ Attempt to fill state space model parameters """
        a = self.a if a is None else a
        Fs = self.Fs if Fs is None else Fs
        if w is not None and Fs is not None:
            if freq is not None:
                assert (freq == OscillatorModel.rad_to_hz(w, Fs)).all(), 'Inconsistent input frequency parameters.'
            self.w = w
            self.freq = OscillatorModel.rad_to_hz(w, Fs)
        freq = self.freq if freq is None else freq
        if freq is not None and Fs is not None:
            self.w = OscillatorModel.hz_to_rad(freq, Fs)
        sigma2 = self.sigma2 if sigma2 is None else sigma2
        F, Q = self._osc_to_ssm_param(a=a, w=self.w, sigma2=sigma2)
        self.F = F if F is not None else self.F
        self.Q = Q if Q is not None else self.Q

    def _ssm_to_osc_param(self, F=None, Q=None):
        """ Convert from matrices F and Q to oscillator parameters """
        F = self.F if F is None else F
        Q = self.Q if Q is None else Q
        ncomp = self.ncomp
        a = np.zeros(ncomp, dtype=np.float64) if ncomp > 0 else None
        w = np.zeros(ncomp, dtype=np.float64) if ncomp > 0 else None
        sigma2 = np.zeros(ncomp, dtype=np.float64) if ncomp > 0 else None
        for ii in range(ncomp):
            a[ii], w[ii] = OscillatorModel.get_rot_param(
                F[ii * 2:ii * 2 + 2, ii * 2:ii * 2 + 2]) if F is not None else (None, None)
            sigma2[ii] = Q[ii * 2, ii * 2] if Q is not None else None  # type: ignore
        return a, w, sigma2

    def _osc_to_ssm_param(self, a=None, w=None, sigma2=None):
        """ Convert from oscillator parameters to matrices F and Q """
        # a, w, sigma2 must be index-able
        a = self.a if a is None else a
        w = self.w if w is None else w
        sigma2 = self.sigma2 if sigma2 is None else sigma2
        if w is not None:
            F_blocks = []
            Q_blocks = []
            for ii in range(w.size):
                F_blocks.append(OscillatorModel.get_rot_mat(a[ii], w[ii]))
                Q_blocks.append(np.eye(2, dtype=np.float64) * sigma2[ii])
            return block_diag(*F_blocks), block_diag(*Q_blocks)
        else:
            return None, None

    def get_default_q(self, components=None, E=None):
        """
        Get the default structure of state noise covariance
        matrix Q in the Q_basis block diagonal form
        """
        components = self.components if components is None else components
        if len(components) == 1 or isinstance(components, OscillatorModel):
            E = np.eye(1, dtype=np.float64) if E is None else E
            default_Q = block_diag(E, E)
        else:
            default_Q = block_diag(*[x.get_default_q(components=x, E=E) for x in components])
        return default_Q

    @staticmethod
    def tr(B):
        """ Compute the trace in a 2x2 matrix """
        assert B.shape == (2, 2), 'B is not a 2x2 matrix.'
        return B[0, 0] + B[1, 1]

    @staticmethod
    def rt(B):
        """ Compute the off-diagonal entry difference in a 2x2 matrix """
        assert B.shape == (2, 2), 'B is not a 2x2 matrix.'
        return B[1, 0] - B[0, 1]

    @staticmethod
    def get_rot_mat(a, w):
        """ F = a * R(w) where R is a rotation matrix with rotation radian w """
        F = np.zeros((2, 2), dtype=np.float64)
        F[0, 0] = a * cos(w)
        F[0, 1] = a * -sin(w)
        F[1, 0] = a * sin(w)
        F[1, 1] = a * cos(w)
        return F

    @staticmethod
    def get_rot_param(F):
        """ F = a * R(w) where R is a rotation matrix with rotation radian w """
        assert F.shape == (2, 2), 'F is not a 2x2 matrix.'
        a = sqrt(F[0, 0] ** 2 + F[1, 0] ** 2)
        w = atan2(F[1, 0], F[0, 0])
        return a, w

    @staticmethod
    def hz_to_rad(frequency_hz, sampling_frequency):
        """ Convert rotation frequency from hertz to radian """
        return 2 * np.pi * frequency_hz / sampling_frequency

    @staticmethod
    def rad_to_hz(frequency_rad, sampling_frequency):
        """ Convert rotation frequency from radian to hertz """
        return frequency_rad * sampling_frequency / (2 * np.pi)

    # Visualization methods - Author: Amanda Beck <ambeck@mit.edu>
    def whiten(self, filter_type='sos'):
        """
        Adapted from Purdon and Weisskoff, Human Brain Mapping (1998), fit dc component and then
        whiten by inverting a fitted AR1 (the dc component)
        """
        # Make sure the dc component is fitted. This is why we allow only one component
        assert self.dc_idx is not None, 'DC Component is necessary for whitening. dc_idx is None.'
        assert self.ncomp == 1, 'More than one component. Only whiten with a single dc component.'

        # Calculate the whitening filter
        # noinspection PyProtectedMember
        from .._preprocessing._colored_noise_utils import _spectral_factorization, _setup_filters
        sos1, sos2, (D, gamma, K) = _spectral_factorization(self.R, self.sigma2, self.a)
        # noinspection PyAttributeOutsideInit
        self.whitener = dict(D=D, gamma=gamma, K=(D, gamma, K))

        # Apply whitening filter with LCCDE
        if filter_type == 'LCCDE':
            self.y_whiten = np.concatenate((self.y[0], np.zeros(self.y.size - 1)))
            for ii in range(self.y.size - 1):
                self.y_whiten[0, ii + 1] = np.sqrt(self.whitener['K']) * (
                        self.y[0, ii + 1] - self.whitener['gamma'] * self.y[0, ii]) + self.a * self.y_whiten[0, ii]
        if filter_type == 'sos':
            whitener, colorer = _setup_filters(sos1, sos2)
            # noinspection PyAttributeOutsideInit
            self.y_whiten = whitener(self.y)

    def visualize_freq(self, version, bw=1, y=None, sim_osc=None, sim_x=None, xlim=None, ylim=None, ax=None):
        """ Visualize the frequency spectrum of real data or the theoretical PSD of the oscillation components """
        import matplotlib.pyplot as plt
        from ..multitaper import fast_psd_multitaper

        # using '%matplotlib notebook' allows zooming in some contexts
        # Create an axes handle for the plot if needed
        if ax is None:
            _, ax = plt.subplots(1, 1)
        fig = ax.get_figure()
        colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        # colors = ['tab:blue', 'tab:orange', 'tab:green', 'tab:red', 'tab:purple', 'tab:pink', 'tab:olive', 'tab:cyan']

        # Plot observed data spectra
        y = self.y if y is None else y
        y = self._must_be_row(self._process_constructor_input(y))  # make sure y is a 2D row vector
        if y is not None:
            psd_mt, fy_hz = fast_psd_multitaper(y, self.Fs, 0, self.Fs / 2, bw)
            ax.plot(fy_hz.squeeze(), 10 * np.log10(psd_mt.squeeze()), color='black', label='observed data')
        if self.y_whiten is not None:
            # noinspection PyUnboundLocalVariable
            ax.plot(fy_hz, 10 * np.log10(psd_mt.squeeze() / psd_mt.min()), linestyle='dashed', color='black',
                    label='scaled_data')
            psd_mt, fy_hz = fast_psd_multitaper(self.y_whiten.squeeze(), self.Fs, 0, self.Fs / 2, bw)
            ax.plot(fy_hz, 10 * np.log10(psd_mt.squeeze()), label='whitened_data')

        # Plot oscillator spectra
        if version == 'actual':
            kalman_out = self.dejong_filt_smooth(y, return_dict=True)
            h_i, f_hz = self.oscillator_spectra(version, kalman_out['x_t_n'][:, 1:], bw)
        else:
            h_i, f_hz = self.oscillator_spectra(version)

        h_sum = np.zeros((1, h_i[0].size))
        for ii in range(len(h_i)):
            ax.plot(f_hz, 10 * np.log10(h_i[ii]), label='osc %d, %0.2f Hz' % (ii + 1, np.float64(self.freq[ii])))
            h_sum += np.sqrt(h_i[ii])

        if version == 'theoretical':
            ax.axhline(10 * np.log10(2 * self.R), color='black', linestyle='dashed', label='obs noise')  # one-sided
        else:
            ax.plot(f_hz, 20 * np.log10(h_sum.squeeze()), color='gray', label='sum of components')

        # Handle simulated oscillator spectra plotting
        if sim_osc is not None and (sim_x is not None or version == 'theoretical'):
            h_sim, f_sim = sim_osc.oscillator_spectra(version, sim_x, bw)

            for ii in range(len(h_sim)):
                ax.plot(f_sim, 10 * np.log10(h_sim[ii]), linestyle='dashed', color=colors[ii],
                        label='sim osc %d' % (ii + 1))

        elif sim_osc is None and sim_x is not None:
            raise RuntimeError('Do you want to plot empirical spectra of simulated data? If yes,'
                               'input sim_osc object in addition to latent states sim_x.')
        elif sim_osc is not None and sim_x is None:
            raise RuntimeError('Do you want to plot empirical spectra of simulated data? If yes,'
                               'input latent states as sim_x.')

        # Final axes adjustment
        ax.legend()
        ax.set_xlabel('Frequency (Hz)')
        ax.set_ylabel('PSD (dB)')
        if xlim is not None:
            ax.set_xlim(xlim)
        if ylim is not None:
            ax.set_ylim(ylim)
        fig.set_tight_layout(True)

        return fig

    def visualize_time(self, y=None, plot_ospe=False, ospe_ylim=None, sim_x=None, xlim=None, figsize=(8, 8)):
        """
        Visualize time series from fitted oscillations.
        sim_x is a 2 dim array (2*ncomp, T) containing the simulated latent states
        """
        import matplotlib.pyplot as plt
        colors = ['tab:blue', 'tab:orange', 'tab:green', 'tab:red', 'tab:purple', 'tab:pink', 'tab:olive', 'tab:cyan']

        y = self.y if y is None else y
        y = self._must_be_row(self._process_constructor_input(y))  # make sure y is a 2D row vector
        kalman_out = self.dejong_filt_smooth(y, return_dict=True)
        x_est = kalman_out['x_t_n'][0::2, 1:]
        y_est = np.sum(x_est, axis=0)
        ta = np.arange(y_est.size) / self.Fs

        subplot_total = self.ncomp + 1
        if plot_ospe:
            # subplot_total += 1
            x_pred = kalman_out['x_t_tmin1'][:, 1:]
            y_pred = self.G @ x_pred
            ospe = (y - y_pred).squeeze()

        fig, axs = plt.subplots(subplot_total, 1, sharex='all', figsize=figsize) \
            if plot_ospe is False else plt.subplots(subplot_total + 1, 1, sharex='all', figsize=figsize)

        # Plot observed and estimated total signal (y)
        axs[0].plot(ta, y.squeeze(), color='black')
        axs[0].plot(ta, y_est.squeeze(), color='gray')
        axs[0].legend(['observed data (true y)', 'estimated y'], loc='lower left')
        if hasattr(self, 'y_whiten') and self.y_whiten is not None:
            axs[0].plot(ta, self.y_whiten.squeeze())
            axs[0].legend(['real data', 'estimated y', 'whitened y'], loc='lower left')
        if xlim is not None:
            axs[0].set_xlim(xlim)
        axs[0].set_title('Total Signal')
        ax0_lim = axs[0].get_ylim()

        # Plot each oscillation
        for ii in range(1, subplot_total):
            if sim_x is not None:
                if ii <= int(sim_x.shape[0] / 2):
                    axs[ii].plot(ta, sim_x[2 * (ii - 1), :].squeeze(), color='gray', label='simulated oscillation')
            axs[ii].plot(ta, x_est[ii - 1, :], color=colors[ii - 1], label='estimated oscillation')
            axs[ii].set_ylim(ax0_lim)
            if xlim is not None:
                axs[ii].set_xlim(xlim)
            axs[ii].legend(loc='lower left')
            axs[ii].set_title('Oscillation %d' % ii)

        # Plot One-Step Prediction Error
        if plot_ospe:
            # noinspection PyUnboundLocalVariable
            axs[subplot_total].plot(ta, ospe, color='black', label='OSPE')
            axs[subplot_total].set_ylim(ax0_lim)
            if ospe_ylim is not None:
                axs[subplot_total].set_ylim(ospe_ylim)

            if xlim is not None:
                axs[subplot_total].set_xlim(xlim)
            axs[subplot_total].legend(loc='lower left')
            axs[subplot_total].set_title('One-Step Prediction Error')
            axs[subplot_total].set_xlabel('Time (sec)')
        else:
            axs[subplot_total - 1].set_xlabel('Time (sec)')

        fig.set_tight_layout(True)

        return fig, x_est

    def oscillator_spectra(self, version='theoretical', x_matrix=None, bw=1):
        """
        Compute theoretical or empirical/actual oscillator spectra
        - Note: these spectra are one-sided from 0 Hz to Nyquist frequency,
        and these spectra correspond to PSD estimates normalized by the
        number of time points but not normalized by sampling frequency.
        """
        from ..multitaper import fast_psd_multitaper

        if version == 'theoretical':
            h_i, f_hz = self._theoretical_spectrum()
        elif version == 'actual':
            h_i = []
            for ii in range(self.ncomp):
                # note that this is not set up for dc components
                psd_mt, f_hz = fast_psd_multitaper(x_matrix[ii * 2, :], self.Fs, 0, self.Fs / 2, bw)
                h_i.append(psd_mt)
        else:
            raise ValueError('Chosen version is not supported. Please select theoretical or actual.')

        # noinspection PyUnboundLocalVariable
        return h_i, f_hz

    def _theoretical_spectrum(self, f_hz=None):
        """ Compute spectrum based on theoretical equation (Matsuda 2017) """
        f_hz = np.linspace(0, self.Fs / 2, 2**12) if f_hz is None else f_hz  # one-sided frequency 0 to Nyquist
        rads = f_hz * 2 * np.pi / self.Fs
        z = np.exp(1j * rads)
        a_i = (1 - 2 * self.a ** 2 * np.cos(self.w) ** 2 + self.a ** 4 * np.cos(2 * self.w)) / (
                self.a * (self.a ** 2 - 1) * np.cos(self.w))
        b_i = 0.5 * (a_i - 2 * self.a * np.cos(self.w) + np.sqrt((a_i - 2 * self.a * np.cos(self.w)) ** 2 - 4))
        v_i = -self.sigma2 * self.a * np.cos(self.w) / b_i
        h_i = [2 * v_ii * np.abs(1 + b_ii * z) ** 2 / np.abs(1 - 2 * a_ii * np.cos(w_ii) * z + a_ii ** 2 * z ** 2) ** 2
               for v_ii, b_ii, a_ii, w_ii in zip(v_i, b_i, self.a, self.w)]  # one-sided spectrum
        return h_i, f_hz

    def _theoretical_phase(self):
        """ Compute phase response based on the complex spectrum """
        rads = np.linspace(-np.pi, np.pi, 1000)
        z = np.exp(-1j * rads)
        a_i = (1 - 2 * self.a ** 2 * np.cos(self.w) ** 2 + self.a ** 4 * np.cos(2 * self.w)) / (
                self.a * (self.a ** 2 - 1) * np.cos(self.w))
        b_i = 0.5 * (a_i - 2 * self.a * np.cos(self.w) + np.sqrt((a_i - 2 * self.a * np.cos(self.w)) ** 2 - 4))
        complex_base_spec = [(1 + b_ii * z) / (1 - 2 * a_ii * np.cos(w_ii) * z + a_ii ** 2 * z ** 2) for
                             b_ii, a_ii, w_ii
                             in zip(b_i, self.a, self.w)]
        phase = [np.arctan2(np.imag(spec), np.real(spec)) for spec in complex_base_spec]
        return phase, rads, complex_base_spec

    # Parameter estimation methods (M step)
    def m_estimate(self, priors=None, **kwargs):
        """
        Maximum likelihood or Maximum a posteriori estimation to update
        parameters. Calling super().m_estimate() for all Ssm parameters
        """
        # Handle DC oscillators if present
        if self.dc_idx is not None:
            if priors is None:
                priors: list = [None] * self.ncomp
                priors[self.dc_idx] = {'dc': True}
            elif isinstance(priors, dict):
                priors['dc'] = True
                priors = [priors]
            else:
                priors[self.dc_idx]['dc'] = True

        results = super().m_estimate(priors=priors, **kwargs)
        self.update_comp_param()
        return results

    def update_comp_param(self):
        """ Update OscillatorModel specific parameters """
        self.fill_osc_param()

    def initialize_priors(self, kappa=None, f_prior=None, Q_sigma2=None, Q_hyperparameter=None,
                          R_sigma2=None, R_hyperparameter=None):
        """
        Initialize priors for Matsuda oscillators

        - INPUT RULE: if None is given or no explicit argument is provided, default values
        will be initialized and used as prior parameters. To initialize priors only for
        some parameters but not others, use the input value 'MLE' on desired parameters.
        """
        # base case
        if self.ncomp <= 1:
            # [Von Mises prior] on rotation frequency w
            if f_prior is None:
                _, w_prior = OscillatorModel.get_rot_param(self.F)
                f_prior = OscillatorModel.rad_to_hz(w_prior, self.Fs)
            else:
                w_prior = OscillatorModel.hz_to_rad(f_prior, self.Fs)
            kappa = 10000. if kappa is None else kappa
            vmp_param = {'kappa': kappa, 'f_prior': f_prior, 'w_prior': w_prior}

            # [Inverse gamma prior] on state noise covariance Q diagonal entries
            if Q_sigma2 is None:
                assert self.Q[0, 0] == self.Q[1, 1], 'State noise sigma2 differs between real and imaginary parts.'
                Q_sigma2 = self.Q[0, 0]
                Q_hyperparameter = 0.1 if Q_hyperparameter is None else Q_hyperparameter

            # [Inverse gamma prior] on observation noise variance R <--- TODO: update to Wishart
            if R_sigma2 is None:
                if self.R.shape[0] > 1:
                    raise NotImplementedError('Only uni-variate observation data is supported with prior for now.')
                else:
                    R_sigma2 = self.R[0, 0]
                    R_hyperparameter = 0.1 if R_hyperparameter is None else R_hyperparameter

            return {'vmp_param': vmp_param, 'Q_sigma2': Q_sigma2, 'Q_hyperparameter': Q_hyperparameter,
                    'R_sigma2': R_sigma2, 'R_hyperparameter': R_hyperparameter}

        # recursive case
        else:
            assert self.components is not None, 'Cannot initialize priors outside base case when components is None.'
            components_prefill = self.fill_components()

            # expand the specified prior values to the length of components
            f_prior = self._initialize_priors_recursive_list(f_prior)
            kappa = self._initialize_priors_recursive_list(kappa)
            Q_sigma2 = self._initialize_priors_recursive_list(Q_sigma2)
            R_sigma2 = self._initialize_priors_recursive_list(R_sigma2)
            R_hyperparameter = self._initialize_priors_recursive_list(R_hyperparameter)
            Q_hyperparameter = self._initialize_priors_recursive_list(Q_hyperparameter)

            # construct the final priors that is a list of dictionaries
            priors = []
            for ii in range(self.ncomp):
                current_component: OscillatorModel = self.components[ii]
                assert current_component.type == 'osc', 'Component type is not OscillatorModel class.'

                priors.append(current_component.initialize_priors(kappa=kappa[ii], f_prior=f_prior[ii],
                                                                  Q_sigma2=Q_sigma2[ii], R_sigma2=R_sigma2[ii],
                                                                  R_hyperparameter=R_hyperparameter[ii],
                                                                  Q_hyperparameter=Q_hyperparameter[ii]))

            # unfill the components
            self.unfill_components(components_prefill)

            return priors

    @staticmethod
    def ig_lp(mode, alpha, estimate):
        """ Compute inverse gamma log prior based on parameters (ignoring terms that don't depend on estimate) """
        return -(alpha + 1) * np.log(estimate) - mode * (alpha + 1) / estimate

    @staticmethod
    def vmp_lp(vmp_dict, estimate):
        """ Compute von mises prior based on parameters (ignoring terms that don't depend on estimate) """
        return vmp_dict['kappa'] * np.cos(estimate - vmp_dict['w_prior'])

    @staticmethod
    def _m_update_f(A=None, B=None, priors=None):
        """ Update transition matrix -- F """
        # Update the rotation radian -- w (adaptive Von Mises prior)
        if Ssm._m_update_if_mle(['dc', 'vmp_param'], priors):
            # MLE
            w_new = atan2(OscillatorModel.rt(B),
                          OscillatorModel.tr(B))
        elif 'dc' in priors:
            # fix DC oscillator frequency at 0
            w_new = 0.
        elif 'vmp_param' in priors:
            # MAP with Von Mises prior
            kappa = priors['vmp_param']['kappa']  # effective kappa is kappa * a_new/sigma2_Q_new
            w_prior = priors['vmp_param']['w_prior']
            w_new = atan2(OscillatorModel.rt(B) + kappa * sin(w_prior),
                          OscillatorModel.tr(B) + kappa * cos(w_prior))
        else:
            raise RuntimeError('Could not set m_update rule for rotation radian w. Please check.')

        # Update the rotation damping factor -- a (no prior)
        a_new = (cos(w_new) * OscillatorModel.tr(B) + sin(w_new) * OscillatorModel.rt(B)) / OscillatorModel.tr(A)

        # Construct transition matrix -- F
        F = OscillatorModel.get_rot_mat(a_new, w_new)
        return F

    @staticmethod
    def _m_update_q(A=None, B=None, C=None, T=None, F=None, priors=None):
        """ Update state noise covariance matrix -- Q """
        a_new, w_new = OscillatorModel.get_rot_param(F)
        Q_ss = (OscillatorModel.tr(C) - 2 * a_new * (cos(w_new) * OscillatorModel.tr(B) +
                                                     sin(w_new) * OscillatorModel.rt(B))
                + a_new ** 2 * OscillatorModel.tr(A))

        T *= 2  # double the data length since we sum over real and imaginary noise processes

        if Ssm._m_update_if_mle('Q_sigma2', priors):
            # MLE
            sigma2_Q_new = Q_ss / T
        else:
            # MAP with inverse gamma prior
            Q_init = priors['Q_sigma2']
            Q_hp = priors['Q_hyperparameter'] if 'Q_hyperparameter' in priors else 0.1
            alpha = T * Q_hp / 2  # scales with data length T according to the hyperparameter
            beta = Q_init * (alpha + 1)  # setting the mode of inverse gamma prior to be Q_init
            sigma2_Q_new = (beta + Q_ss / 2) / (alpha + T / 2 + 1)  # mode of inverse gamma posterior

        Q = sigma2_Q_new * np.eye(2, dtype=sigma2_Q_new.dtype)
        return Q

    @staticmethod
    def _m_update_q_src(Q_ss=None, T=None, Q_basis=None, nsource=None, nstate=None, priors=None):
        """
        Update state noise covariance matrix Q in dynamic
        source localization with a single Q_basis, in the
        Q_basis block diagonal form
        """
        import torch
        from ..source_loc import SourceLocModel as Src
        assert nstate == 2, 'OscillatorModel must have a component state dimension of 2.'
        assert Q_ss.shape[0] == nsource * nstate, 'Q_ss does not match the dimension of OscillatorModel' \
                                                  ' across ' + str(nsource) + ' sources.'

        # Combine sums of squares across diagonal blocks
        Q_ss_block = torch.zeros((nsource, nsource), dtype=Q_ss.dtype).cuda()
        for j in range(nstate):
            start_idx = j * nsource
            end_idx = (j + 1) * nsource
            Q_ss_block += Q_ss[start_idx:end_idx, start_idx:end_idx]

        # Update the expansion coefficient for each source
        if isinstance(Q_basis, list):  # using non-orthonormal kernel
            Theta = Src.update_theta(Q_ss=Q_ss_block, T=T, Q_basis=Q_basis[1],
                                     nsource=nsource, npart=nstate, priors=priors)
            Q_new = torch.block_diag(*[Q_basis[0] @ torch.diag(Theta) @ Q_basis[0].T] * nstate)
        else:  # using orthonormal basis
            Theta = Src.update_theta(Q_ss=Q_ss_block, T=T, Q_basis=Q_basis,
                                     nsource=nsource, npart=nstate, priors=priors)
            Q_new = torch.block_diag(*[Q_basis @ torch.diag(Theta) @ Q_basis.T] * nstate)
        return Q_new

    @staticmethod
    def _m_update_S0(x_0_n=None, P_0_n=None, mu0=None):
        """ Update initial state covariance -- S0 """
        sigma2_S0 = OscillatorModel.tr(P_0_n + x_0_n[:, None] @ x_0_n[:, None].T
                                       - x_0_n[:, None] @ mu0.T - mu0 @ x_0_n[:, None].T + mu0 @ mu0.T) / 2
        S0 = sigma2_S0 * np.eye(2, dtype=sigma2_S0.dtype)
        return S0

    @staticmethod
    def _m_update_g(y=None, x_t_n=None, P_t_n=None, h_t=None, C=None, D=None):
        """ Update observation matrix -- G (OscillatorModel has fixed G) """
        return None



"""
Author: Amanda Beck <ambeck@mit.edu>
        Mingjian He <mh1@stanford.edu>
"""

from .helper_functions import (
    np, initialize_newosc, plot_innovations,
    get_innovations, get_knee, aic_calc, fit_ar
    )
from ..basic_models import OscillatorModel as Osc
from .decomp_osc import DecomposedOscillatorModel
from ..utils import estimate_r


class IterativeOscillatorModel(DecomposedOscillatorModel):
    """
    IterativeOscillatorModel is an object class containing fitted OscillatorModel objects
    from the iterative oscillator search algorithm
    """
    def __init__(self, y, fs, noise_start=None, osc_range=7, iterate=False, **kwargs):
        """
        Inputs:
            :param y: observed data
            :param fs: sampling frequency (Hz)
            :param noise_start: frequency (Hz) above which white noise is assumed to estimate R
            :param osc_range: maximum number of oscillators
            :param iterate: shortcut to apply the method iterate() on the initialized IterativeOscillatorModel
            :param **kwargs: key-word pair arguments for iterate()
        """
        if noise_start is None:
            noise_start = fs / 2 - 20
            if noise_start <= 0:
                noise_start = fs / 2 - 5
        assert noise_start > 0, 'Please redefine noise_start. Entered or calculated frequency is less than zero.'

        # initialize the first oscillator
        add_freq, add_radius, add_sigma2, *_ = initialize_newosc(fs, y)
        R = estimate_r(y, fs, freq_cutoff=noise_start)
        init_params = {'y': y, 'a': add_radius, 'freq': add_freq, 'sigma2': add_sigma2, 'Fs': fs, 'R': R}

        # populate instance attributes
        self.noise_start = noise_start
        self.osc_range = osc_range
        self.freq_res = None
        self.keep_param = ()

        self.added_osc = [Osc(**init_params)]
        self.fitted_osc = []
        self.priors = []
        self.all_params = []  # a list of lists containing Osc objects with only parameters (y dropped)
        self.ll = []
        self.ll_xre = []
        self.AIC = []
        self.knee_index = None
        self.reiterated = False

        self.iterate(**kwargs) if iterate else None  # perform learning iterations

    def iterate(self, freq_res=1, keep_param=(), reiterate=False, sigma2_method='eig',
                freq_hp=0, R_hp=0.1, Q_hp=None, R_sigma2='constant', Q_sigma2='MLE',
                no_priors=False, track_params=False, plot_fit=False, verbose=False):
        """
        Iterative Oscillator Search (iOsc) Algorithm

        Reference:
            Beck, A. M., He, M., Gutierrez, R. G., & Purdon, P. L. (2022). An iterative
            search algorithm to identify oscillatory dynamics in neurophysiological
            time series. bioRxiv, 2022-10.

            Matsuda & Komaki (2017). Time Series Decomposition into Oscillation
            Components and Phase Estimation. Journal of Neural Computation, 29, 332-367.

            Shumway, R. H., & Stoffer, D. S. (1982). An approach to time series
            smoothing and forecasting using the EM algorithm. Journal of time series
            analysis, 3(4), 253-264.

        Inputs:
        :param self: IterativeOscillatorModel class instance
        :param freq_res: minimal oscillator frequency spacing/resolution when adding a new oscillator
        :param keep_param: a tuple of strings for parameters to keep and not update
        :param reiterate: whether to perform reiterations by sorting log-likelihood increases in descending order
        :param sigma2_method: method used to initialize sigma2 of oscillators
        :param freq_hp: concentration hyperparameter in Von Mises Prior (freq_hp * data length)
        :param R_hp: hyperparameter in Inverse Gamma Prior determining weight on prior mode compared to MLE
        :param Q_hp: hyperparameter in Inverse Gamma Prior determining weight on prior mode compared to MLE
        :param R_sigma2: determine prior mode for R, if None it will be noise variance from previous iteration
        :param Q_sigma2: determine prior mode for Q, if None all components use noise variances from previous iteration
        :param no_priors: override all other prior hyperparameters and proceed with no priors
        :param track_params: save parameters from EM iterations
        :param plot_fit: plot the fitted oscillators and innovations at each iteration
        :param verbose: print oscillator information during iterative search iterations
        """
        # Store iteration settings as instance attributes
        self.freq_res = freq_res
        self.keep_param = keep_param

        # Copy the first added oscillator to preserve its parameters
        o1 = self.added_osc[0].copy()

        # Set up iteration variables
        start_add_osc = False
        innov_limit = None
        counter = 0

        # Begin iterations to search for more oscillators in the observed data
        while o1.ncomp < self.osc_range:
            if not start_add_osc:  # do EM learning on the initial oscillator first, don't add oscillators yet
                start_add_osc = True
                if verbose:
                    print('EM learning on the initial oscillator')

            else:  # add an oscillator
                # find existing frequency
                if abs(o1.freq[-1] - self.added_osc[-1].freq) >= self.freq_res:
                    existing_freqs = list(o1.freq) + [x.freq[0] for x in self.added_osc[:-1]]
                else:
                    existing_freqs = list(o1.freq) + [x.freq[0] for x in self.added_osc]

                add_freq, add_radius, add_sigma2, _, _, a, p = initialize_newosc(
                    o1.Fs, innovations, existing_freqs=np.abs(existing_freqs), freq_res=self.freq_res,  # noqa: F821
                    ar_order=self.osc_range * 2 - 1, sigma2=sigma2_method)

                if verbose:
                    print('AR: add freq = %0.2f Hz, add radius = %0.3f, add sigma2 = %0.3f'
                          % (add_freq, add_radius, add_sigma2))

                # plot the fitted oscillator from previous iteration and the innovation
                if plot_fit:
                    _, innov_limit = plot_innovations(o1, o1.y, innovations, a, p, add_freq, add_radius,  # noqa: F821
                                                      ax_limit=innov_limit, counter=counter)

                # construct an additional oscillator
                o2 = Osc(a=add_radius, freq=add_freq, sigma2=add_sigma2, Fs=o1.Fs)
                self.added_osc.append(o2)
                o1.append(o2)  # add the additional oscillator to the existing oscillator instance
                counter += 1  # accumulate iteration counter

            # Initialize priors for EM learning
            if no_priors:
                priors = None
            else:  # provide some informative priors to constrain parameter updates
                prior_params = {'R_hyperparameter': R_hp, 'Q_hyperparameter': Q_hp,
                                'R_sigma2': R_sigma2, 'Q_sigma2': Q_sigma2}
                if R_sigma2 == 'constant':  # use the same prior mode throughout iterations
                    prior_params['R_sigma2'] = self.added_osc[0].R[0, 0]
                if verbose:
                    print('Prior input parameters:', prior_params)
                priors = o1.initialize_priors(kappa=o1.ntime * freq_hp, **prior_params)
            self.priors.append(priors)  # store the priors dictionary

            # Run EM iterations
            em_params = []  # store the oscillator parameters throughout EM iterations
            for _ in range(50):
                _ = o1.m_estimate(**o1.dejong_filt_smooth(EM=True), priors=priors, keep_param=self.keep_param)
                if track_params:
                    em_params.append(o1.copy(drop_y=True))
            if track_params:
                self.all_params.append(em_params)

            self.fitted_osc.append(o1.copy(drop_y=True))

            if verbose:
                print('Iteration %d completed:' % counter)
                print(o1)

            # Find the innovation spectrum
            innovations, ll = get_innovations(o1)
            self.ll.append(ll)
            self.AIC.append(aic_calc(o1, ll))  # compute AIC

        if reiterate:
            # Perform reiterations
            self.reiterate(track_params=track_params, plot_fit=plot_fit, verbose=verbose)
        else:
            # Find the knee_index for the selected model of fitted oscillators
            self.knee_index = get_knee(self.ll)

    def reiterate(self, track_params=False, plot_fit=False, verbose=False):
        """ Sort and re-add the oscillators in descending order of log-likelihood increase """
        # Fast-forward to after fitting the first oscillator
        o1 = self.fitted_osc[0].copy()
        o1.y = self.added_osc[0].y.copy()  # add the observed data back in

        # Fill in the instance attributes for the first fitted oscillator
        added_osc_re = self.added_osc[:1]
        fitted_osc_re = self.fitted_osc[:1]
        priors_re = self.priors[:1]
        all_params_re = self.all_params[:1]
        innovations, ll = get_innovations(o1)
        ll_re = [ll]
        AIC_re = [aic_calc(o1, ll)]

        # Set up iteration variables
        add_indices = np.diff(self.ll).argsort()[::-1] + 1
        innov_limit = None
        counter = 0

        # Reiterate through all added oscillators
        for add_idx in add_indices:
            # construct the oscillator to be added next
            add_radius = self.fitted_osc[add_idx].a[-1]
            add_freq = np.abs(self.fitted_osc[add_idx].freq[-1])
            add_sigma2 = self.fitted_osc[add_idx].sigma2[-1]
            o2 = Osc(a=add_radius, freq=add_freq, sigma2=add_sigma2, Fs=o1.Fs)

            if verbose:
                print('Re-iteration: add freq = %0.2f Hz, add radius = %0.3f, add sigma2 = %0.3f'
                      % (o2.freq[0], o2.a[0], o2.sigma2[0]))

            # plot the fitted oscillator from previous iteration and the innovation
            if plot_fit:
                a, p = fit_ar(np.squeeze(innovations), self.osc_range * 2 - 1)
                _, innov_limit = plot_innovations(o1, o1.y, innovations, a, p, o2.freq[0], o2.a[0], reiterate=True,
                                                  ax_limit=innov_limit, counter=counter)

            added_osc_re.append(o2)
            o1.append(o2)  # add the additional oscillator to the existing oscillator instance
            counter += 1  # accumulate iteration counter

            # Initialize priors for EM learning
            if self.priors[add_idx] is None:  # then no_priors was True during iterate()
                priors = None
            else:
                prior_params = self.priors[add_idx][0].copy()
                kappa = prior_params['vmp_param']['kappa']
                del prior_params['vmp_param']
                if verbose:
                    print('Prior input parameters:', prior_params)
                priors = o1.initialize_priors(kappa=kappa, **prior_params)
            priors_re.append(priors)

            # Run EM iterations
            em_params = []  # store the oscillator parameters throughout EM iterations
            for x in range(50):
                _ = o1.m_estimate(**o1.dejong_filt_smooth(EM=True), priors=priors, keep_param=self.keep_param)
                if track_params:
                    em_params.append(o1.copy(drop_y=True))
            if track_params:
                all_params_re.append(em_params)

            fitted_osc_re.append(o1.copy(drop_y=True))

            if verbose:
                print('Re-iteration %d completed:' % counter)
                print(o1)

            # Find the innovation spectrum
            innovations, ll = get_innovations(o1)
            ll_re.append(ll)
            AIC_re.append(aic_calc(o1, ll))  # compute AIC

        # Overwrite the instance attributes from iterate()
        self.added_osc = added_osc_re
        self.fitted_osc = fitted_osc_re
        self.priors = priors_re
        self.all_params = all_params_re
        self.ll_xre = self.ll  # save the log likelihoods before reiterate()
        self.ll = ll_re  # overwrite the log likelihoods after reiterate()
        self.AIC = AIC_re

        # Find the new knee_index for the selected model of re-fitted oscillators
        self.knee_index = get_knee(self.ll)  # knee replacement

        # Update the reiterated flag
        self.reiterated = True

    def __repr__(self):
        """ Unambiguous and concise representation when calling IterativeOscillatorModel() """
        return 'IterOsc(' + str(len(self)) + ')<' + hex(id(self))[-4:] + '>'



"""
Author: Mingjian He <mh1@stanford.edu>

source_loc module contains dynamic source localization code base in SOMATA
"""

from .dmap_em import SourceLocModel  # Source localization model class  # noqa: F401
